#include <ADQAPI.h>

#include <vector>
#include <map>
#include <set>

unsigned int g_max_nof_channels = 0;
unsigned int g_max_bytes_per_sample = 0;
std::vector<ADQDaisyChainDeviceInformation> g_device_info;

void ThrowException(unsigned int status, const char* ErrorSource, bool ThrowIt=true) {
	if (status != 1)
	{
		if (ThrowIt)
			throw ErrorSource;
		else
			printf("%s had an error\n", ErrorSource);
	}
}

unsigned long long GetElapsedTime(ULARGE_INTEGER const& StartTime)
{

	FILETIME NowTimeF;
	GetSystemTimeAsFileTime(&NowTimeF);
	ULARGE_INTEGER NowTime;
	NowTime.HighPart = NowTimeF.dwHighDateTime;
	NowTime.LowPart = NowTimeF.dwLowDateTime;

	return NowTime.QuadPart - StartTime.QuadPart;
}

bool Master(std::string const& SerialNumber)
{
	bool Master = false;
	if ((SerialNumber == "SPD-06905") || (SerialNumber == "SPD-06911"))
		Master = true;
	return Master;
}

bool Master(ADQInterface* adqInterface)
{
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	return Master(SerialNumber);
}

void SetTriggerIdleState(ADQInterface* adqInterface, bool ThrowIt = true)
{
	ThrowException(adqInterface->SetInternalTriggerSyncMode(0), "SetInternalTriggerSyncMode", ThrowIt);
	ThrowException(adqInterface->DisarmTimestampSync(), "DisarmTimestampSync", ThrowIt);
	ThrowException(adqInterface->DisarmTrigger(), "DisarmTrigger", ThrowIt);
	ThrowException(adqInterface->SetPreTrigSamples(0), "SetPreTrigSamples", ThrowIt);
	ThrowException(adqInterface->SetTriggerDelay(0), "SetTriggerDelay", ThrowIt);
}

void SetDaisyChainIdleState(ADQInterface* adqInterface, bool ThrowIt = true)
{
	ThrowException(adqInterface->DaisyChainEnable(0), "DaisyChainEnable", ThrowIt);
	ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset", ThrowIt);
}

void Initialise(void* adqCtrlUnit, int adqDevListNum,
	int nof_samples, std::vector<ADQInterface*>& adqInterfaces)
{
	// Opens communication channel to a certain ADQ device
	ThrowException(ADQControlUnit_OpenDeviceInterface(adqCtrlUnit, adqDevListNum), "ADQControlUnit_OpenDeviceInterface");

	// Make this device ready to use
	ThrowException(ADQControlUnit_SetupDevice(adqCtrlUnit, adqDevListNum), "ADQControlUnit_SetupDevice");

	ADQInterface* adqInterface = ADQControlUnit_GetADQ(adqCtrlUnit, adqDevListNum + 1);
	adqInterfaces[adqDevListNum] = adqInterface;
	ThrowException(adqInterface ? 1 : 0, "ADQControlUnit_GetADQ");

	ThrowException(IS_VALID_DLL(adqInterfaces[adqDevListNum]), "IS_VALID_DLL");

	int* FWRevision = adqInterfaces[adqDevListNum]->GetRevision();
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	printf("Serial number %s FW Revision %d\n", SerialNumber.c_str(), *FWRevision);

	g_max_nof_channels = __max(g_max_nof_channels, adqInterfaces[adqDevListNum]->GetNofChannels());

	unsigned int bytes_per_sample;
	ThrowException(adqInterface->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
	g_max_bytes_per_sample = __max(g_max_bytes_per_sample, bytes_per_sample);

	SetTriggerIdleState(adqInterface);
	SetDaisyChainIdleState(adqInterface);

	ThrowException(adqInterfaces[adqDevListNum]->SetClockSource(3), "SetClockSource"); // Internal clock source, external 10 MHz reference from PXIsync
}

unsigned int DaisyChainGetStatus(ADQInterface* adqInterface)
{
	unsigned int DaisyChainStatus = 0;
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	ThrowException(adqInterface->DaisyChainGetStatus(&DaisyChainStatus), "DaisyChainGetStatus");
	if (DaisyChainStatus)
		printf("WARNING: Daisy chain status %s %u\n", SerialNumber.c_str(), DaisyChainStatus);
	return DaisyChainStatus;
}

void SetDaisyChain(ADQInterface* adqInterface)
{
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	bool Master = ::Master(SerialNumber);
	ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset");
	if (Master)
	{
		ThrowException(adqInterface->DaisyChainSetMode(1), "DaisyChainSetMode"); // Master mode
		ThrowException(adqInterface->DaisyChainSetupOutput(1, 0, 12), "DaisyChainSetupOutput");
	}
	else
	{
		ThrowException(adqInterface->DaisyChainSetMode(0), "DaisyChainSetMode"); // Slave mode
		ThrowException(adqInterface->DaisyChainSetupOutput(1, 0, 12), "DaisyChainSetupOutput");
	}
	ThrowException(adqInterface->DaisyChainEnableOutput(1), "DaisyChainEnableOutput");
	ThrowException(adqInterface->DaisyChainEnable(1), "DaisyChainEnable");
	DaisyChainGetStatus(adqInterface);
}

void SetInputRange(ADQInterface* adqInterface)
{
	bool TerminatingModule = false;
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	if (SerialNumber == "SPD-06916")
		TerminatingModule = true;

	for (int Channel = 0; Channel < 8; Channel++)
	{
		float result;
		ThrowException(adqInterface->SetInputRange(Channel + 1, 1000, &result), "SetInputRange");
		unsigned int termination = 1; // default high impedance.
		if ((TerminatingModule) && (Channel == 1))
			termination = 0;
		ThrowException(adqInterface->SetInputImpedance(Channel + 1, termination), "SetInputImpedance"); // High impedance
		ThrowException(adqInterface->SetAdjustableBias(Channel + 1, 0), "SetAdjustableBias");
	}
}

void SetTriggerMode(ADQInterface* adqInterface, double Pre_Trig, unsigned int Position, bool HardTrig)
{
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	bool Master = ::Master(SerialNumber);
	ThrowException(adqInterface->DisarmTrigger(), "DisarmTrigger");
	double sampleratehz;
	ThrowException(adqInterface->GetSampleRate(0, &sampleratehz), "GetSampleRate");
	int PretriggerSamples = 0; // int(Pre_Trig * sampleratehz + 0.5);
	ThrowException(adqInterface->SetClockSource(3), "SetClockSource"); // Internal clock source, external 10 MHz reference from PXIsync
	if (Master)
	{
		if (HardTrig)
		{
			ThrowException(adqInterface->SetTriggerMode(3), "SetTriggerMode"); // Level trigger mode
			ThrowException(adqInterface->SetTriggerEdge(3, 1), "SetTriggerEdge");
			ThrowException(adqInterface->SetLvlTrigLevel(1024), "SetLvlTrigLevel");
			const int ChannelH = 7;
			ThrowException(adqInterface->SetLvlTrigChannel(1 << ChannelH), "SetLvlTrigChannel"); // Channel "H"

			// ThrowException(adqInterface->SetupTimestampSync(0, 3), "SetupTimestampSync"); Can't use the level trigger for synch.
			ThrowException(adqInterface->DaisyChainSetTriggerSource(3), "DaisyChainSetTriggerSource");
			ThrowException(adqInterface->DaisyChainSetupLevelTrigger(ChannelH + 1, 1024, 1, 1), "DaisyChainSetupLevelTrigger");
		}
		else
		{
			ThrowException(adqInterface->SetTriggerMode(4), "SetTriggerMode"); // Internal trigger mode
			ThrowException(adqInterface->SetTriggerEdge(4, 1), "SetTriggerEdge");
			ThrowException(adqInterface->SetupTimestampSync(0, 4), "SetupTimestampSync");
			ThrowException(adqInterface->DaisyChainSetTriggerSource(4), "DaisyChainSetTriggerSource");
		}
		ThrowException(adqInterface->SetInternalTriggerPeriod(UINT_MAX - 4), "SetInternalTriggerPeriod");
	}
	else
	{
		ThrowException(adqInterface->SetTriggerMode(23), "SetTriggerMode"); // daisy chain trigger mode
		ThrowException(adqInterface->SetInternalTriggerSyncMode(2), "SetInternalTriggerSyncMode"); // Slave mode
		ThrowException(adqInterface->SetupTimestampSync(0, 9), "SetupTimestampSync"); //Get an error for 23!
	}
	unsigned int nof_pretrigger_samples;
	ThrowException(adqInterface->DaisyChainGetNofPretriggerSamples(Position, 1, &nof_pretrigger_samples), "DaisyChainGetNofPretriggerSamples");
	unsigned int TriggerDelaySamples = int(-Pre_Trig * sampleratehz + 0.5);
	TriggerDelaySamples = 4 * (TriggerDelaySamples / 4); // Must be a multiple of 4.
	PretriggerSamples += nof_pretrigger_samples;
	/*	if (PretriggerSamples >= 0)
		{
			TriggerDelaySamples = 0;
		}
		else
		{
			TriggerDelaySamples = -PretriggerSamples;
			TriggerDelaySamples = 4 * (TriggerDelaySamples / 4); // Must be a multiple of 4.
			PretriggerSamples = 0;
		}*/
	ThrowException(adqInterface->SetSampleSkip(1), "SetSampleSkip");
	ThrowException(adqInterface->SetPreTrigSamples(PretriggerSamples), "SetPreTrigSamples");
	ThrowException(adqInterface->SetTriggerDelay(TriggerDelaySamples), "SetTriggerDelay");
}


unsigned int IndexToPosition(int adqDevListNum)
{
	unsigned int Position = adqDevListNum + 2 * ((adqDevListNum + 1) % 2) - 1;
	return Position;
}

std::map<unsigned int, std::map<unsigned int, unsigned int>> SlaveMapping(std::vector<ADQInterface*> const& adqInterfaces)
{
	std::map<unsigned int, std::map<unsigned int, unsigned int>> SlaveMapping;
	std::vector<bool> Master(adqInterfaces.size());
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		Master[adqDevListNum] = ::Master(adqInterfaces[adqDevListNum]);
		if (Master[adqDevListNum])
		{
			SlaveMapping[adqDevListNum] = std::map<unsigned int, unsigned int>();
		}
	}
	unsigned int LastMaster = SlaveMapping.begin()->first;
	unsigned int Offset = 0;
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		if ((adqDevListNum > 1) && (adqDevListNum + 1 < adqInterfaces.size()) && (Master[adqDevListNum+1]))
		{
			SlaveMapping[LastMaster][adqDevListNum - 1] = 2 * (SlaveMapping[LastMaster][adqDevListNum - 1] / 2);
			Offset = adqDevListNum;
			LastMaster = adqDevListNum + 1;
		}
		if (!Master[adqDevListNum])
		{
			unsigned int Position = IndexToPosition(adqDevListNum - Offset) - IndexToPosition(LastMaster - Offset);
			SlaveMapping[LastMaster][adqDevListNum] = Position;
		}
	}
	if (SlaveMapping.size() == 0)
		ThrowException(0, "No masters set");
	SlaveMapping[LastMaster][unsigned(adqInterfaces.size() - 1)] = 2 * (SlaveMapping[LastMaster][unsigned(adqInterfaces.size() - 1)] / 2);
	return SlaveMapping;
}

int DaisyChainGetNofPretriggerSamples(ADQInterface* adqInterface, unsigned int Position, int& TriggerDelaySamples)
{
	const double Pre_Trig = -2.5e-07;
	double sampleratehz;
	ThrowException(adqInterface->GetSampleRate(0, &sampleratehz), "GetSampleRate");
	unsigned int DaisyPretriggerSamples;
	ThrowException(adqInterface->DaisyChainGetNofPretriggerSamples(Position, 1, &DaisyPretriggerSamples), "DaisyChainGetNofPretriggerSamples");
	int PretriggerSamples = int(Pre_Trig * sampleratehz + 0.5);
	if (PretriggerSamples >= 0)
	{
		DaisyPretriggerSamples += PretriggerSamples;
		TriggerDelaySamples = 0;
	}
	else
	{
		TriggerDelaySamples = -PretriggerSamples;
		TriggerDelaySamples = 4 * (TriggerDelaySamples / 4); // Must be a multiple of 4.
	}
//	printf("Position %d, DaisyPretriggerSamples %3d, TriggerDelaySamples %d\n", Position, DaisyPretriggerSamples, TriggerDelaySamples);
	return DaisyPretriggerSamples;
}

uint64_t GetDaisyChainTimeStamps(ADQInterface* adqInterface, unsigned int nof_records, unsigned int nof_fetched_records, unsigned int nof_samples,
	std::map<unsigned int, std::map<unsigned int, unsigned int>>::const_iterator MasterIter,
	std::vector<unsigned long long>& daisy_timestamp, bool HardTrig)
{
	std::vector<ADQDaisyChainDeviceInformation> ADQDaisyChainDeviceInformation;
	int TriggerDelaySamples;
	ADQDaisyChainDeviceInformation.resize(MasterIter->second.size());
	size_t SlaveNo = 0;
	for (std::map<unsigned int, unsigned int>::const_iterator SlaveIter = MasterIter->second.begin();
		SlaveIter != MasterIter->second.end(); SlaveIter++)
	{
		unsigned int Position = SlaveIter->second;
		ADQDaisyChainDeviceInformation[SlaveNo].Position = Position;
		int PretriggerSamples = DaisyChainGetNofPretriggerSamples(adqInterface, Position, TriggerDelaySamples);
		ADQDaisyChainDeviceInformation[SlaveNo].PretriggerSamples = PretriggerSamples;
		ADQDaisyChainDeviceInformation[SlaveNo].TriggerDelaySamples = TriggerDelaySamples;
		SlaveNo++;
	}
	static std::vector<ADQDaisyChainTriggerInformation> ADQDaisyChainTriggerInformation;
	ADQDaisyChainTriggerInformation.reserve(nof_records);
	ADQDaisyChainTriggerInformation.resize(nof_fetched_records + 1);
	static std::vector<std::vector<int64_t>> RecordStart;
	RecordStart.reserve(nof_records);
	RecordStart.resize(ADQDaisyChainTriggerInformation.size());
	static std::vector<double> ExtendedPrecision;
	ExtendedPrecision.resize(MasterIter->second.size());
	for (size_t Record = 0; Record < ADQDaisyChainTriggerInformation.size(); Record++)
	{
		RecordStart[Record].resize(MasterIter->second.size());
		ADQDaisyChainTriggerInformation[Record].RecordStart = &(RecordStart[Record][0]);
		ADQDaisyChainTriggerInformation[Record].ExtendedPrecision = &(ExtendedPrecision[0]);
	}


	ThrowException(adqInterface->DaisyChainGetTriggerInformation(4, 1, 0, 8, unsigned int(ADQDaisyChainTriggerInformation.size()), nof_samples, 1, &(ADQDaisyChainDeviceInformation[0]),
		unsigned int(MasterIter->second.size()), &(ADQDaisyChainTriggerInformation[0])), "DaisyChainGetTriggerInformation");
	printf("Module %s Daisy record starts ", adqInterface->GetBoardSerialNumber());
	for (size_t Module = 0; Module < RecordStart.back().size(); Module++)
		printf("%lld ", RecordStart.back()[Module]);
	printf("\n");

	unsigned int MinPosition = UINT_MAX;
	for (std::map<unsigned int, unsigned int>::const_iterator SlaveIter = MasterIter->second.begin();
		SlaveIter != MasterIter->second.end(); SlaveIter++)
		MinPosition = __min(MinPosition, SlaveIter->second);
	for (std::map<unsigned int, unsigned int>::const_iterator SlaveIter = MasterIter->second.begin();
		SlaveIter != MasterIter->second.end(); SlaveIter++)
		daisy_timestamp[SlaveIter->first] = ADQDaisyChainTriggerInformation.back().Timestamp + RecordStart.back()[SlaveIter->second-MinPosition];
	return ADQDaisyChainTriggerInformation.back().Timestamp+TriggerDelaySamples;
}

void SetPrePostTrigger(ADQInterface* adqInterface, unsigned int Position)
{
	int TriggerDelaySamples;
	int DaisyPretriggerSamples = DaisyChainGetNofPretriggerSamples(adqInterface, Position, TriggerDelaySamples);
	ThrowException(adqInterface->SetPreTrigSamples(DaisyPretriggerSamples), "SetPreTrigSamples");
	ThrowException(adqInterface->SetTriggerDelay(TriggerDelaySamples), "SetTriggerDelay");
}

void AcquireSequence(std::vector<ADQInterface*>& adqInterfaces,
	unsigned int nof_samples, unsigned int nof_records, double TriggerPeriod, bool HardTrig)
{
	std::vector<std::vector<unsigned long long>>
		local_timestamp(nof_records),
		daisy_timestamp(nof_records);
	for (size_t Record = 0; Record < nof_records; Record++)
	{
		local_timestamp[Record].resize(adqInterfaces.size());
		daisy_timestamp[Record].resize(adqInterfaces.size());
	}

	fflush(stdout);

	std::vector<void*> target_buffers(g_max_nof_channels);
	for (unsigned int channel = 0; channel < target_buffers.size(); channel++)
		target_buffers[channel] = new unsigned char[nof_samples*g_max_bytes_per_sample];

	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
		ThrowException(adqInterfaces[adqDevListNum]->MultiRecordSetup(nof_records, nof_samples), "MultiRecordSetup");

	std::vector<bool> Master(adqInterfaces.size());

	std::map<unsigned int, std::map<unsigned int, unsigned int>> SlaveMapping = ::SlaveMapping(adqInterfaces);
	unsigned int LastMaster = SlaveMapping.begin()->first;
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		Master[adqDevListNum] = ::Master(adqInterfaces[adqDevListNum]);
		if (Master[adqDevListNum])
			SetPrePostTrigger(adqInterfaces[adqDevListNum], 0);
	}
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		if (adqDevListNum > 0)
			// This Sleep just to make it obvious whether the trigger stream starts with - or without - using the daisy chain.
			Sleep(100);
		if ((adqDevListNum > 1) && (adqDevListNum + 1 < adqInterfaces.size()) && (Master[adqDevListNum + 1]))
		{
			LastMaster = adqDevListNum + 1;
		}
		if (!Master[adqDevListNum])
		{
			if ((adqDevListNum > 1) && (adqDevListNum + 1 < adqInterfaces.size()) && (Master[adqDevListNum + 1]))
			{
				LastMaster = adqDevListNum + 1;
			}
			std::map<unsigned int, std::map<unsigned int, unsigned int>>::const_iterator MasterIter =
				SlaveMapping.find(LastMaster);
			if (MasterIter == SlaveMapping.end())
				ThrowException(0, "Slavemapping");
			std::map<unsigned int, unsigned int>::const_iterator SlaveIter = MasterIter->second.find(adqDevListNum);
			if (SlaveIter == MasterIter->second.end())
				ThrowException(0, "Slavemapping");
			unsigned int Position = SlaveIter->second;
			SetPrePostTrigger(adqInterfaces[adqDevListNum], Position);
			ThrowException(adqInterfaces[adqDevListNum]->ArmTimestampSync(), "ArmTimestampSync");
			ThrowException(adqInterfaces[adqDevListNum]->ArmTrigger(), "ArmTrigger");
		}
	}
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		if (Master[adqDevListNum])
		{
			double sampleratehz;
			ThrowException(adqInterfaces[adqDevListNum]->GetSampleRate(0, &sampleratehz), "GetSampleRate");
			unsigned int internTrigPeriod = unsigned int(0.5 + TriggerPeriod * sampleratehz);
			ThrowException(adqInterfaces[adqDevListNum]->SetInternalTriggerPeriod(internTrigPeriod), "SetInternalTriggerPeriod");
			unsigned int internTrigHighSamp = internTrigPeriod / 2,
				internTrigLowSamp = internTrigPeriod / 2;
			ThrowException(adqInterfaces[adqDevListNum]->SetInternalTriggerHighLow(internTrigHighSamp, internTrigLowSamp), "SetInternalTriggerHighLow");
			ThrowException(adqInterfaces[adqDevListNum]->SetTriggerEdge(4, 1), "SetTriggerEdge");
			ThrowException(adqInterfaces[adqDevListNum]->ArmTimestampSync(), "ArmTimestampSync");
			ThrowException(adqInterfaces[adqDevListNum]->ArmTrigger(), "ArmTrigger");
		}
	}

	std::vector<unsigned int> nof_fetched_records(adqInterfaces.size());
	unsigned char ch_mask = 0xFF; // ch A+B+C+D+E+F+G+H

	FILETIME StartTimeF;
	GetSystemTimeAsFileTime(&StartTimeF);
	ULARGE_INTEGER StartTime;
	StartTime.HighPart = StartTimeF.dwHighDateTime;
	StartTime.LowPart = StartTimeF.dwLowDateTime;
	std::vector<ULARGE_INTEGER> LastElapsedTime(adqInterfaces.size());

	unsigned int min_nof_fetched_records = 0;
	unsigned int nof_reported_records = 0;
	std::vector<bool> DaisyChainStatusLogged(adqInterfaces.size());
	std::vector<bool> StreamOverflowLogged(adqInterfaces.size());
	std::vector<unsigned int> nof_acquired_records(adqInterfaces.size());
	while (min_nof_fetched_records < nof_records)
	{
		min_nof_fetched_records = UINT_MAX;
		bool GotAtLeastOne = false;
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
		{
			ADQInterface* adqInterface = adqInterfaces[adqDevListNum];
			nof_acquired_records[adqDevListNum] = adqInterface->GetAcquiredRecords();
			if (nof_acquired_records[adqDevListNum] > nof_fetched_records[adqDevListNum])
			{
				/* Transfer Data from ADQ to host - records one by one */
				/* Re-use target buffer allocated on each call */
				unsigned int bytes_per_sample;
				ThrowException(adqInterface->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
				unsigned int Record = nof_fetched_records[adqDevListNum];
				ThrowException(adqInterface->GetDataWHTS(&target_buffers[0], NULL, &(local_timestamp[Record][adqDevListNum]), nof_samples, bytes_per_sample,
					Record, 1, ch_mask, 0, nof_samples, ADQ_TRANSFER_MODE_NORMAL), "GetDataWHTS");
				if (Master[adqDevListNum])
				{
					std::map<unsigned int, std::map<unsigned int, unsigned int>>::const_iterator MasterIter =
						SlaveMapping.find(adqDevListNum);
					if (MasterIter == SlaveMapping.end())
						ThrowException(0, "Slavemapping");
					daisy_timestamp[Record][adqDevListNum] = GetDaisyChainTimeStamps(adqInterface, nof_records, Record, nof_samples, MasterIter, daisy_timestamp[Record], HardTrig);
				}
				nof_fetched_records[adqDevListNum]++;
				GotAtLeastOne = true;
				StreamOverflowLogged[adqDevListNum] = false;
				DaisyChainStatusLogged[adqDevListNum] = false;
			}
			else
			{
				// Frequently logs warnings?
				if ((!StreamOverflowLogged[adqDevListNum]) && (adqInterface->GetStreamOverflow()))
				{
					printf("WARNING: Stream overflow on %s\n", adqInterface->GetBoardSerialNumber());
					StreamOverflowLogged[adqDevListNum] = true;
				}
				// Sometimes logs warnings?
				if ((!DaisyChainStatusLogged[adqDevListNum]) && (DaisyChainGetStatus(adqInterface)))
					DaisyChainStatusLogged[adqDevListNum] = true;
			}
			min_nof_fetched_records = __min(min_nof_fetched_records, nof_fetched_records[adqDevListNum]);
		}
		if (!GotAtLeastOne)
		{
			Sleep(1); // Wait some before polling again
		}
		else if (min_nof_fetched_records > nof_reported_records)
		{
			printf("record # % 0d\n", nof_reported_records);
			for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
			{
				unsigned long long local_tdiff = local_timestamp[nof_reported_records][adqDevListNum];
				if (nof_reported_records > 0)
					local_tdiff -= local_timestamp[nof_reported_records - 1][adqDevListNum];
				unsigned long long daisy_tdiff = daisy_timestamp[nof_reported_records][adqDevListNum];
				if (nof_reported_records > 0)
					daisy_tdiff -= daisy_timestamp[nof_reported_records - 1][adqDevListNum];
				double local_tdiff_d = (double)local_tdiff * 0.000000000025;
				double daisy_tdiff_d = (double)daisy_tdiff * 0.000000000025;
				unsigned long long ElapsedTime = GetElapsedTime(StartTime);
				double DiffElapsed = (ElapsedTime - LastElapsedTime[adqDevListNum].QuadPart) / 1E7; // 100ns intervals.
				printf("Module %2d (%s) acquired: %4d records, elapsed time is %8llu (diff = %5.4f seconds), local timestamp is %14llu (diff = %5.4f seconds), daisy timestamp is %14llu (diff = %5.4f seconds)\n",
					adqDevListNum, Master[adqDevListNum] ? "Master" : "Slave ", nof_acquired_records[adqDevListNum], ElapsedTime, DiffElapsed, local_timestamp[nof_reported_records][adqDevListNum], local_tdiff_d, daisy_timestamp[nof_reported_records][adqDevListNum], daisy_tdiff_d);
				LastElapsedTime[adqDevListNum].QuadPart = ElapsedTime;
				if (nof_reported_records == 0)
				{
					// Anomaly: the first timestamp for a internally-triggered (master) record does not get synchronised to 0.
					if (Master[adqDevListNum])
						local_timestamp[0][adqDevListNum] = 0;
					daisy_timestamp[0][adqDevListNum] = 0;
				}
			}
			nof_reported_records++;
		}
	}
	for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
	{
		/** Close MultiRecord **/
		adqInterfaces[adqDevListNum]->DisarmTrigger();
		adqInterfaces[adqDevListNum]->DisarmTimestampSync();
		if (Master[adqDevListNum])
			adqInterfaces[adqDevListNum]->SetInternalTriggerPeriod(UINT_MAX - 4);
		adqInterfaces[adqDevListNum]->MultiRecordClose();
	}
	for (unsigned int channel = 0; channel < target_buffers.size(); channel++)
	{
		delete[] target_buffers[channel];
	}
}

int main(int argc, const char* argv[])
{
	void* adqCtrlUnit = NULL;
	double TriggerPeriod = 0.5; // in sec
	bool HardTrig = false;
	std::vector<unsigned int> nof_records(1, 4);

	if (argc > 1)
		TriggerPeriod = atof(argv[1]);

	if (argc > 2)
	{
		HardTrig = (atoi(argv[2]) != 0);
	}

	if (argc > 3)
	{
		nof_records.resize(argc - 3);
		for (size_t Sequence = 0; Sequence < argc - 3; Sequence++)
			nof_records[Sequence] = atoi(argv[Sequence + 3]);
	}

	timeBeginPeriod(1);
	std::vector<ADQInterface*> adqInterfaces;

	unsigned int adqDevList = 0;
	unsigned int nof_samples = 700;
	const double Pre_Trig = -2.5e-07;
	unsigned int Group = 1;

	/** Set Up Collection with MultiRecord **/
	try {
		adqCtrlUnit = CreateADQControlUnit();
		ThrowException(adqCtrlUnit != NULL, "CreateADQControlUnit");

		ThrowException(ADQControlUnit_EnableErrorTrace(adqCtrlUnit, LOG_LEVEL_INFO, "."), "ADQControlUnit_EnableErrorTrace");

		struct ADQInfoListEntry* adqInfoStruct;

		ThrowException(ADQControlUnit_ListDevices(adqCtrlUnit, &adqInfoStruct, &adqDevList), "ADQControlUnit_ListDevices");

		adqInterfaces.resize(adqDevList);
		std::vector<unsigned int> nof_channels(adqInterfaces.size());

		int APIRevision = ADQAPI_GetRevision();
		printf("API Revision %d\n", APIRevision);

		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
			Initialise(adqCtrlUnit, adqDevListNum, nof_samples, adqInterfaces);
		unsigned int LastMaster = 0;
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
			SetInputRange(adqInterfaces[adqDevListNum]);
		for (size_t Sequence = 0; Sequence < nof_records.size(); Sequence++)
		{
			bool ThisHardTrig = (Sequence == nof_records.size() - 1) ? HardTrig : false;
			for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
			{
				if (Master(adqInterfaces[adqDevListNum]))
					LastMaster = adqDevListNum;
				int Position = adqDevListNum - LastMaster;
				SetTriggerMode(adqInterfaces[adqDevListNum], Pre_Trig, Position / Group, ThisHardTrig);
			}

			printf("\nAcquiring: %000d records at %f.\n", nof_records[Sequence], TriggerPeriod);
			AcquireSequence(adqInterfaces, nof_samples, nof_records[Sequence], TriggerPeriod, ThisHardTrig);
		}
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqInterfaces.size(); adqDevListNum++)
			SetDaisyChainIdleState(adqInterfaces[adqDevListNum], false);
	}
	catch(const char* ErrorSource)
	{
		printf("%s had an error\n", ErrorSource);
	}
	if (adqCtrlUnit)
	{
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
			ThrowException(ADQControlUnit_CloseDevice(adqCtrlUnit, adqDevListNum), "ADQControlUnit_CloseDevice", false);
		DeleteADQControlUnit(adqCtrlUnit);
	}
	timeEndPeriod(1);
	return 0;
}
