#include <ADQAPI.h>

#include <vector>
#include <map>

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	class GetWHTSTest																	//
//	Description:																		//
//		A C++ wrapper round the test implementation.									//
//		The test aims to investigate this things:										//
//			a. Use of the daisy chain.													//
//			b. Daisy chain trigger delay measurement.									//
//			c. Use of both internal and external triggering.							//								
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
class GetWHTSTest
{
public:
	GetWHTSTest(unsigned int nof_samples, unsigned int Group);
	void Initialise();
	ADQInterface* Initialise(void* adqCtrlUnit, int adqDevListNum);
	~GetWHTSTest();
	const unsigned int internTrigHighSamp = 500;
	void ThrowException(unsigned int status, const char* ErrorSource, bool ThrowIt = true);
	unsigned long long GetElapsedTime(ULARGE_INTEGER const& StartTime);
	void SetDaisyChain();
	unsigned int DaisyChainGetStatus(ADQInterface* adqInterface, long FromLine);
	void SetTriggerIdleState(ADQInterface* adqInterface, bool ThrowIt = true);
	void SetDaisyChainIdleState(ADQInterface* adqInterface, bool ThrowIt = true);
	void detect_daisy_chain_order();
	void SetDaisyChain(ADQInterface* adqInterface, unsigned int Position);
	void SetInputRange(ADQInterface* adqInterface);
	void SetTriggerMode(double Pre_Trig, bool HardTrig);
	void SetTriggerMode(ADQInterface* adqInterface, double Pre_Trig, unsigned int Position, bool HardTrig);
	void WaitForAcquisitionComplete(unsigned int nof_records, ULARGE_INTEGER StartTime);
	void AcquireSequence(
		unsigned int nof_records, double TriggerPeriod, bool HardTrig);

	std::vector<ADQInterface*> adqInterfaces() {
		return m_adqInterfaces;
	}

private:
	unsigned long long DaisyChainGetTriggerInformation2(unsigned int adqDevListNum, unsigned int nof_fetched_records);
	unsigned int m_max_nof_channels;		// The maximum number of channels available for any module. This is always 8 for ADQ8
	unsigned int m_max_bytes_per_sample;	// The maximum number of bytes used by any module. This is always 2 for ADQ8.
	void* m_adqCtrlUnit;					// The ADQ control unit object.
	unsigned int m_adqDevList;				// The number of devices detected.
	std::map<int, std::vector<int>> m_device_position;
	std::vector<ADQInterface*> m_adqInterfaces;					// The list of ADQ objects
	std::vector<ADQDaisyChainDeviceInformation> m_device_info;	// This information is provided to 
	std::vector<unsigned long long> m_TimeStampOffsetOfPRE;
	std::vector<unsigned long long> m_TimeStamp;
	std::map<int, unsigned long long> m_DaisyTimeStampOfPRE;
	std::map<int, std::vector<int64_t>> m_RecordStartOfPRE;
	std::map<std::string, FILE*> m_CSVFiles;
	unsigned int m_nof_samples;
	unsigned int m_Group;
};

static const int ChannelH = 7;			// The trigger input channel (counted from 0).
static const int TriggerLevel = 1024;	// The trigger level.

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	GetWHTSTest::GetWHTSTest(unsigned int nof_samples, unsigned int Group)				//
//	GetWHTSTest::~GetWHTSTest()															//
//																						//
//	Description:																		//
//		Constructor and destructor.														//
//		nof_samples will be collected per record.										//
//		Group selects how the daisy chain is grouped - with or without grouping modules	//
//		togther by means of the sync_immediate parameter of DaisyChainSetupOutput.		//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
GetWHTSTest::GetWHTSTest(unsigned int nof_samples, unsigned int Group)
{
	m_max_nof_channels = 0;
	m_max_bytes_per_sample = 0;
	m_nof_samples = nof_samples;
	m_Group = Group;
	m_adqCtrlUnit = NULL;
	m_adqDevList = NULL;
}

GetWHTSTest::~GetWHTSTest()
{
	if (m_adqCtrlUnit)
	{
		for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
			if (m_adqInterfaces[adqDevListNum])
				SetDaisyChainIdleState(m_adqInterfaces[adqDevListNum], false);
		for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqDevList; adqDevListNum++)
			ThrowException(ADQControlUnit_CloseDevice(m_adqCtrlUnit, adqDevListNum), "ADQControlUnit_CloseDevice", false);
		DeleteADQControlUnit(m_adqCtrlUnit);
	}
	for (std::map<std::string, FILE*>::iterator CSVIter = m_CSVFiles.begin(); CSVIter != m_CSVFiles.end(); CSVIter++)
		fclose(CSVIter->second);
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::ThrowException(unsigned int status, const char* ErrorSource,		//
//		bool ThrowIt /*= true*/)														//
//																						//
//	Description:																		//
//		Throws ErrorSource if the status is bad.										//
//		Optionally, logs a warning instead, if ThrowIt is false.						//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::ThrowException(unsigned int status, const char* ErrorSource, bool ThrowIt /*= true*/)
{
	if (status != 1)
	{
		if (ThrowIt)
			throw ErrorSource;
		else
			printf("WARN : %s had an error\n", ErrorSource);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	unsigned long long GetWHTSTest::GetElapsedTime(ULARGE_INTEGER const& StartTime)		//
//																						//
//	Description:																		//
//		Returns the precision time elapsed since StartTime								//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
unsigned long long GetWHTSTest::GetElapsedTime(ULARGE_INTEGER const& StartTime)
{

	FILETIME NowTimeF;
	GetSystemTimeAsFileTime(&NowTimeF);
	ULARGE_INTEGER NowTime;
	NowTime.HighPart = NowTimeF.dwHighDateTime;
	NowTime.LowPart = NowTimeF.dwLowDateTime;

	return NowTime.QuadPart - StartTime.QuadPart;
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::SetTriggerIdleState(ADQInterface* adqInterface,					//
//		bool ThrowIt /*= true*/)														//
//																						//
//	Description:																		//
//		Set the internal trigger hardware into a default and idle state.				//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::SetTriggerIdleState(ADQInterface* adqInterface, bool ThrowIt /*= true*/)
{
	ThrowException(adqInterface->DisarmTrigger(), "DisarmTrigger", ThrowIt);
	ThrowException(adqInterface->SetInternalTriggerPeriod(UINT_MAX - 4), "SetInternalTriggerPeriod", ThrowIt);
	// The trigger period is set to a very long interval - effectivley infinite in order to prevent any inadvertent
	// triggers from ocurring. I don't know that this is neccecary.
	uint32_t internTrigLowSamp = UINT_MAX - 4 - internTrigHighSamp;
	ThrowException(adqInterface->SetInternalTriggerHighLow(internTrigHighSamp, internTrigLowSamp), "SetInternalTriggerHighLow", ThrowIt);
	ThrowException(adqInterface->SetInternalTriggerSyncMode(0), "SetInternalTriggerSyncMode", ThrowIt);
	ThrowException(adqInterface->DisarmTimestampSync(), "DisarmTimestampSync", ThrowIt);
	ThrowException(adqInterface->SetTriggerMode(1), "SetTriggerMode", ThrowIt); // SW trigger
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::SetDaisyChainIdleState(ADQInterface* adqInterface,				//
//		bool ThrowIt /*= true*/)														//
//																						//
//	Description:																		//
//		Sets the daisy chain into a default idle state.									//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::SetDaisyChainIdleState(ADQInterface* adqInterface, bool ThrowIt /*= true*/)
{
	ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset", ThrowIt);
	DaisyChainGetStatus(adqInterface, __LINE__);
	ThrowException(adqInterface->DaisyChainEnable(0), "DaisyChainEnable", ThrowIt);
	DaisyChainGetStatus(adqInterface, __LINE__);
	ThrowException(adqInterface->DaisyChainSetOutputState(0), "DaisyChainSetOutputState");
	DaisyChainGetStatus(adqInterface, __LINE__);
}

void GetWHTSTest::detect_daisy_chain_order()
{
	// First disable all outputs
	for (int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
	{
		ThrowException(m_adqInterfaces[adq_num]->DaisyChainEnableOutput(0), "DaisyChainEnableOutput");
		ThrowException(m_adqInterfaces[adq_num]->DaisyChainSetOutputState(0), "DaisyChainSetOutputState");
	}
	std::vector<int> next_device(m_adqInterfaces.size(), -1);
	std::vector<int> prev_device(m_adqInterfaces.size(), -1);
	for (int adq_out_num = 0; adq_out_num < m_adqInterfaces.size(); adq_out_num++)
	{
		// Determine which digitizer is not pointed to by any device.
		// Filter out the last slave(next_device == -1)
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainEnableOutput(1), "DaisyChainEnableOutput");
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainSetOutputState(1), "DaisyChainSetOutputState");
		for (int adq_in_num = 0; adq_in_num < m_adqInterfaces.size(); adq_in_num++)
		{
			if (adq_in_num == adq_out_num)
				continue;
			unsigned int InputState = 0;
			ThrowException(m_adqInterfaces[adq_in_num]->DaisyChainGetInputState(&InputState), "DaisyChainSetOutputState");
			if (InputState == 1)
			{
				if (next_device[adq_out_num] != -1)
					ThrowException(0, "Device input already connected");
				next_device[adq_out_num] = adq_in_num;
				if (prev_device[adq_in_num] != -1)
					ThrowException(0, "Device output already connected");
				prev_device[adq_in_num] = adq_out_num;
			}
		}
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainSetOutputState(0), "DaisyChainSetOutputState");
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainEnableOutput(0), "DaisyChainEnableOutput");
	}
	m_device_position.clear();
	int nof_unconnected_slaves = 0;
	for (int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
	{
		if (next_device[adq_num] == -1)
			nof_unconnected_slaves++;
		if (prev_device[adq_num] == -1)
			m_device_position[adq_num] = std::vector<int>();
	}
	if (nof_unconnected_slaves != m_device_position.size())
		ThrowException(0, "At least one incomplete chain");

	// Walk through the next_device list to determine the devices' positions in
	// the daisy chain.
	for (std::map<int, std::vector<int>>::iterator DaisyIter = m_device_position.begin();
		DaisyIter != m_device_position.end(); DaisyIter++)
	{
		int adqDevListNum = DaisyIter->first;
		int next_slave_num = DaisyIter->first;
		ADQInterface* adqInterface = m_adqInterfaces[next_slave_num];
		printf("Master %s is %d FW Revision %d\n", adqInterface->GetBoardSerialNumber(), next_slave_num, *adqInterface->GetRevision());
		while (true)
		{
			next_slave_num = next_device[next_slave_num];
			if (next_slave_num == -1)
				break;
			adqInterface = m_adqInterfaces[next_slave_num];
			printf("Slave %s is %d FW Revision %d\n", adqInterface->GetBoardSerialNumber(), next_slave_num, *adqInterface->GetRevision());
			DaisyIter->second.push_back(next_slave_num);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	int GetWHTSTest::Initialise(void* adqCtrlUnit, int adqDevListNum)					//
//	void GetWHTSTest::Initialise()														//
//																						//
//	Description:																		//
//		Initialises one unit for operation.												//
//		The latter method initialises all units.										//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
ADQInterface* GetWHTSTest::Initialise(void* adqCtrlUnit, int adqDevListNum)
{
	// Opens communication channel to a certain ADQ device
	ThrowException(ADQControlUnit_OpenDeviceInterface(adqCtrlUnit, adqDevListNum), "ADQControlUnit_OpenDeviceInterface");
	// Make this device ready to use
	ThrowException(ADQControlUnit_SetupDevice(adqCtrlUnit, adqDevListNum), "ADQControlUnit_SetupDevice");

	ADQInterface* adqInterface = ADQControlUnit_GetADQ(adqCtrlUnit, adqDevListNum + 1);
	ThrowException(adqInterface ? 1 : 0, "ADQControlUnit_GetADQ");

	ThrowException(IS_VALID_DLL(adqInterface), "IS_VALID_DLL");
	ThrowException(adqInterface->IsStartedOK(), "IsStartedOK");

	m_max_nof_channels = __max(m_max_nof_channels, adqInterface->GetNofChannels());

	unsigned int bytes_per_sample;
	ThrowException(adqInterface->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
	m_max_bytes_per_sample = __max(m_max_bytes_per_sample, bytes_per_sample);

	SetTriggerIdleState(adqInterface);
	SetDaisyChainIdleState(adqInterface);
	return adqInterface;
}

void GetWHTSTest::Initialise()
{
	CopyFile("GetDataWHTS.ini", "GetDataWHTS.hioc", FALSE);
	FILE* DataFile = fopen("GetDataWHTS.hioc", "ab");
	fputc(char(26), DataFile);
	fclose(DataFile);
	m_adqCtrlUnit = CreateADQControlUnit();
	ThrowException(m_adqCtrlUnit != NULL, "CreateADQControlUnit");

	ThrowException(ADQControlUnit_EnableErrorTrace(m_adqCtrlUnit, LOG_LEVEL_RTONLY, "."), "ADQControlUnit_EnableErrorTrace");

	struct ADQInfoListEntry* adqInfoStruct;

	ThrowException(ADQControlUnit_ListDevices(m_adqCtrlUnit, &adqInfoStruct, &m_adqDevList), "ADQControlUnit_ListDevices");

	m_adqInterfaces.resize(m_adqDevList);

	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
	{
		m_adqInterfaces[adqDevListNum] = Initialise(m_adqCtrlUnit, adqDevListNum);
		if (m_adqInterfaces[adqDevListNum])
		{
			DaisyChainGetStatus(m_adqInterfaces[adqDevListNum], __LINE__);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	unsigned int GetWHTSTest::DaisyChainGetStatus(ADQInterface* adqInterface,			//
//		long FromLine)																	//
//																						//
//	Descritpion:																		//
//		Logs a warning if the daisy chain status indicates a fault condition.			//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
unsigned int GetWHTSTest::DaisyChainGetStatus(ADQInterface* adqInterface, long FromLine)
{
	unsigned int DaisyChainStatus = 0;
	ThrowException(adqInterface->DaisyChainGetStatus(&DaisyChainStatus), "DaisyChainGetStatus");
	if (DaisyChainStatus)
		printf("WARN : Daisy chain status %s %u  from Line %d\n", adqInterface->GetBoardSerialNumber(), DaisyChainStatus, FromLine);
	return DaisyChainStatus;
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::SetDaisyChain(ADQInterface* adqInterface, unsigned int Position)	//
//	void GetWHTSTest::SetDaisyChain()													//
//																						//
//	Description:																		//
//		Enables the daisy chain for this module, either as a master, or as a slave.		//
//		The second method enables the daisy chain for all modules.						//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::SetDaisyChain(ADQInterface* adqInterface, unsigned int Position)
{
	bool Master = (Position == 0);
	ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset");
	DaisyChainGetStatus(adqInterface, __LINE__);
	ThrowException(adqInterface->DaisyChainEnable(0), "DaisyChainEnable");
	DaisyChainGetStatus(adqInterface, __LINE__);

	unsigned int sync_immediate = ((Position % m_Group) == 0) ? 0 : 1;
	if (Master)
		ThrowException(adqInterface->DaisyChainSetMode(1), "DaisyChainSetMode"); // Master mode
	else
		ThrowException(adqInterface->DaisyChainSetMode(0), "DaisyChainSetMode"); // Slave mode
	ThrowException(adqInterface->DaisyChainSetupOutput(1, sync_immediate, 12), "DaisyChainSetupOutput");
	DaisyChainGetStatus(adqInterface, __LINE__);
	ThrowException(adqInterface->DaisyChainEnableOutput(1), "DaisyChainEnableOutput");
	DaisyChainGetStatus(adqInterface, __LINE__);
}

void GetWHTSTest::SetDaisyChain()
{
	for(std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		SetDaisyChain(m_adqInterfaces[MasterIter->first], 0);
		for (int Slave = 0; Slave < MasterIter->second.size(); Slave++)
			SetDaisyChain(m_adqInterfaces[MasterIter->second[Slave]], Slave+1);
	}
}


//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::SetInputRange(ADQInterface* adqInterface)							//
//																						//
//	Descripton:																			//
//		Sets the input range and termination options for the aquisition.				//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::SetInputRange(ADQInterface* adqInterface)
{
	bool TerminatingModule = false;
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	if (SerialNumber == "SPD-06916")
		// This is the terminating module of the connected trigger and input cabling transmission line.
		// THIS WILL NEED TO BE CHANGED for other hardware setups.
		TerminatingModule = true;

	for (int Channel = 0; Channel < 8; Channel++)
	{
		float result;
		// Input range 0 to 1000 mV.
		ThrowException(adqInterface->SetInputRange(Channel + 1, 1000, &result), "SetInputRange");
		unsigned int termination = 1; // default high impedance.
		if ((TerminatingModule) && (Channel == 1))
			// THIS WILL NEED TO BE CHANGED for other hardware setups.
			termination = 0;
		ThrowException(adqInterface->SetInputImpedance(Channel + 1, termination), "SetInputImpedance"); // High impedance
		ThrowException(adqInterface->SetAdjustableBias(Channel + 1, 0), "SetAdjustableBias");
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::SetTriggerMode(ADQInterface* adqInterface, double Pre_Trig,		//
//		unsigned int Position, bool HardTrig)											//
//																						//
//	Decription:																			//
//		Sets the trigger mode for hardware or software triggering.						//
//		The daisy pre-trigger information is recorded for use by						//
//		DaisyChainGetTriggerInformation2 in m_device_info.								//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::SetTriggerMode(ADQInterface* adqInterface, double Pre_Trig, unsigned int Position, bool HardTrig)
{
	Pre_Trig = 0;
	bool Master = (Position==0);
	ThrowException(adqInterface->DisarmTrigger(), "DisarmTrigger");
	ThrowException(adqInterface->SetClockSource(1), "SetClockSource"); // 	Internal clock source, external 10 MHz reference
	double sampleratehz;
	ThrowException(adqInterface->GetSampleRate(0, &sampleratehz), "GetSampleRate");
	int PretriggerSamples = 0;
	if (Master)
	{
		if (HardTrig)
		{
			ThrowException(adqInterface->SetTriggerMode(23), "SetTriggerMode"); // Daisy trigger mode
			ThrowException(adqInterface->SetTriggerEdge(3, 1), "SetTriggerEdge");
			ThrowException(adqInterface->SetLvlTrigLevel(TriggerLevel), "SetLvlTrigLevel");
			ThrowException(adqInterface->SetLvlTrigChannel(1 << ChannelH), "SetLvlTrigChannel"); // Channel "H"

			ThrowException(adqInterface->DaisyChainSetTriggerSource(3), "DaisyChainSetTriggerSource");
			ThrowException(adqInterface->SetupTimestampSync(0, 9), "SetupTimestampSync"); // Can't use the level trigger for synch.
			ThrowException(adqInterface->DaisyChainSetupLevelTrigger(ChannelH + 1, TriggerLevel, 1, 1), "DaisyChainSetupLevelTrigger");
		}
		else
		{
			ThrowException(adqInterface->SetTriggerMode(23), "SetTriggerMode"); // Daisy trigger mode
			ThrowException(adqInterface->SetTriggerEdge(4, 1), "SetTriggerEdge");
			ThrowException(adqInterface->DaisyChainSetTriggerSource(4), "DaisyChainSetTriggerSource");
			ThrowException(adqInterface->SetupTimestampSync(0, 9), "SetupTimestampSync");
		}
		// To ensure there aren't any unexpected internal triggers. I'm not sure this is neccecary.
		ThrowException(adqInterface->SetInternalTriggerPeriod(UINT_MAX - 4), "SetInternalTriggerPeriod");
	}
	else
	{
		ThrowException(adqInterface->SetTriggerMode(23), "SetTriggerMode"); // daisy chain trigger mode
		ThrowException(adqInterface->SetInternalTriggerSyncMode(2), "SetInternalTriggerSyncMode"); // Slave mode
		ThrowException(adqInterface->SetupTimestampSync(0, 9), "SetupTimestampSync"); //Get an error for 23!
	}
	int TriggerDelaySamples = int(-Pre_Trig * sampleratehz + 0.5);
	if (TriggerDelaySamples < 0)
	{
		PretriggerSamples = -TriggerDelaySamples;
		TriggerDelaySamples = 0;
	}
	TriggerDelaySamples = 4 * (TriggerDelaySamples / 4); // Must be a multiple of 4.

	unsigned int nof_pretrigger_samples;
	ThrowException(adqInterface->DaisyChainGetNofPretriggerSamples(Position, 1, &nof_pretrigger_samples), "DaisyChainGetNofPretriggerSamples");
	PretriggerSamples += 4;
	PretriggerSamples += nof_pretrigger_samples;
	m_device_info.resize(__max(m_device_info.size(), Position+1));
	m_device_info[Position].Position = Position;
	m_device_info[Position].PretriggerSamples = PretriggerSamples;
	m_device_info[Position].TriggerDelaySamples = TriggerDelaySamples;
	ThrowException(adqInterface->SetSampleSkip(1), "SetSampleSkip");
	ThrowException(adqInterface->SetPreTrigSamples(PretriggerSamples), "SetPreTrigSamples");
	ThrowException(adqInterface->SetTriggerDelay(TriggerDelaySamples), "SetTriggerDelay");
}

void GetWHTSTest::SetTriggerMode(double Pre_Trig, bool HardTrig)
{
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		SetTriggerMode(m_adqInterfaces[MasterIter->first], Pre_Trig, 0, HardTrig);
		for (int Slave = 0; Slave < MasterIter->second.size(); Slave++)
			SetTriggerMode(m_adqInterfaces[MasterIter->second[Slave]], Pre_Trig, 1 + (Slave / m_Group), HardTrig);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	unsigned long long GetWHTSTest::DaisyChainGetTriggerInformation2(					//
//		ADQInterface* adqInterface, int TrigSource, unsigned int nof_fetched_records)	//
//																						//
//	Description:																		//
//		Extracts the daisy chain trigger information from the master modules.			//
//		Retunrs the timestamp of the master module.										//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
unsigned long long GetWHTSTest::DaisyChainGetTriggerInformation2(
	unsigned int adqDevListNum, unsigned int nof_fetched_records)
{
	ADQDaisyChainTriggerInformation trig_info;
	std::vector<int64_t> RecordStart(m_device_info.size());
	std::vector<double> ExtendedPrecision(m_device_info.size());
	ADQInterface* adqInterface = m_adqInterfaces[adqDevListNum];
	trig_info.RecordStart = &(RecordStart[0]);
	trig_info.ExtendedPrecision = &(ExtendedPrecision[0]);
	ThrowException(adqInterface->DaisyChainGetTriggerInformation2(
		23, 1, TriggerLevel, ChannelH+1, nof_fetched_records, 1, m_nof_samples, 1, &(m_device_info[0]),
		int(m_device_info.size()), &trig_info), "DaisyChainGetTriggerInformation2");
	std::string CSVFileName = adqInterface->GetBoardSerialNumber() + std::string(".csv");
	std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.find(adqDevListNum);
	if (m_CSVFiles[CSVFileName] == NULL)
	{
		m_CSVFiles[CSVFileName] = fopen(CSVFileName.c_str(), "wt");
		fprintf(m_CSVFiles[CSVFileName], "Record, DaisyTimeStamp, ");
		for (size_t Module = 0; Module < RecordStart.size(); Module++)
			fprintf(m_CSVFiles[CSVFileName], "RecordStart %zu, ", Module);
		for (size_t Module = 0; Module <= MasterIter->second.size(); Module++)
		{
			fprintf(m_CSVFiles[CSVFileName], "Time offset %zu", Module);
			if (Module < MasterIter->second.size())
				fprintf(m_CSVFiles[CSVFileName], ", ");
		}
		fprintf(m_CSVFiles[CSVFileName], "\n");
	}

	if (nof_fetched_records == 0)
	{
		m_DaisyTimeStampOfPRE[adqDevListNum] = trig_info.Timestamp;
		m_RecordStartOfPRE[adqDevListNum] = RecordStart;
	}
	trig_info.Timestamp -= m_DaisyTimeStampOfPRE[adqDevListNum];
	fprintf(m_CSVFiles[CSVFileName], "%d, %llu, ", nof_fetched_records, trig_info.Timestamp);
	for (size_t Module = 0; Module < RecordStart.size(); Module++)
	{
		fprintf(m_CSVFiles[CSVFileName], "%lld", RecordStart[Module]);
		fprintf(m_CSVFiles[CSVFileName], ", ");
	}

	int64_t TimeStampOffset = m_TimeStamp[MasterIter->first] - trig_info.Timestamp;
	int64_t RecordOffset = RecordStart[0] - m_RecordStartOfPRE[MasterIter->first][0];

	if (nof_fetched_records == 0)
	{
		m_TimeStampOffsetOfPRE.resize(m_TimeStamp.size());
		m_TimeStampOffsetOfPRE[adqDevListNum] = TimeStampOffset;
	}
	TimeStampOffset -= m_TimeStampOffsetOfPRE[adqDevListNum];
	TimeStampOffset -= RecordStart[0];
	TimeStampOffset += m_RecordStartOfPRE[MasterIter->first][0];

	fprintf(m_CSVFiles[CSVFileName], "%lld, ", TimeStampOffset);
	for (size_t Module = 0; Module < MasterIter->second.size(); Module++)
	{
		adqDevListNum = MasterIter->second[Module];
		TimeStampOffset = m_TimeStamp[adqDevListNum] - trig_info.Timestamp;
		if (nof_fetched_records == 0)
			m_TimeStampOffsetOfPRE[adqDevListNum] = TimeStampOffset;
		TimeStampOffset -= m_TimeStampOffsetOfPRE[adqDevListNum];
		TimeStampOffset -= RecordStart[Module+1];
		TimeStampOffset += m_RecordStartOfPRE[MasterIter->first][Module+1];
		fprintf(m_CSVFiles[CSVFileName], "%lld", TimeStampOffset);
		if (Module + 1 < MasterIter->second.size())
			fprintf(m_CSVFiles[CSVFileName], ", ");
	}
	fprintf(m_CSVFiles[CSVFileName], "\n");
	return trig_info.Timestamp;
}

void GetWHTSTest::WaitForAcquisitionComplete(unsigned int nof_records, ULARGE_INTEGER StartTime)
{
	bool Triggered = false;
	unsigned int min_nof_acquired_records = 0;
	DWORD LastTrigTime = GetTickCount();
	std::vector<unsigned int> nof_acquired_records(m_adqInterfaces.size());
	while (min_nof_acquired_records < nof_records)
	{
		min_nof_acquired_records = UINT_MAX;
		bool GotAtLeastOne = false;
		for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
		{
			ADQInterface* adqInterface = m_adqInterfaces[adqDevListNum];
			unsigned int new_nof_acquired_records = adqInterface->GetAcquiredRecords();
			if (new_nof_acquired_records > nof_acquired_records[adqDevListNum])
			{
				GotAtLeastOne = true;
				Triggered = true;
				nof_acquired_records[adqDevListNum] = new_nof_acquired_records;
				LastTrigTime = GetTickCount();
			}
			min_nof_acquired_records = __min(min_nof_acquired_records, new_nof_acquired_records);
		}
		if (!GotAtLeastOne)
		{
			Sleep(1);
			if ((Triggered) && (GetTickCount() - LastTrigTime > 2000))
			{
				printf("WARNING: timed out on more triggers after %9llu\n", GetElapsedTime(StartTime));
				break;
			}
		}
	}
//	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
//		ThrowException(m_adqInterfaces[adqDevListNum]->DisarmTrigger(), "DisarmTrigger");
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void GetWHTSTest::AcquireSequence(													//
//		unsigned int nof_records, double TriggerPeriod, bool HardTrig)					//
//																						//
//	Description:																		//
//		Performs a test aquisition sequence, using internal or external triggering.		//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void GetWHTSTest::AcquireSequence(
	unsigned int nof_records, double TriggerPeriod, bool HardTrig)
{
	FILE* DataFile = fopen("GetDataWHTS.hioc", "ab");
	bool Triggered = false;
	DWORD LastTrigTime = GetTickCount();
	std::vector<unsigned long long> first_timestamp(m_adqInterfaces.size()),
		last_timestamp(m_adqInterfaces.size());
	m_TimeStamp.assign(m_adqInterfaces.size(), 0);
	fflush(stdout);

	std::vector<std::vector<void*>> target_buffers(m_adqInterfaces.size());
	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
	{
		target_buffers[adqDevListNum].resize(m_max_nof_channels);
		for (unsigned int channel = 0; channel < target_buffers[adqDevListNum].size(); channel++)
			target_buffers[adqDevListNum][channel] = new unsigned char[m_nof_samples*m_max_bytes_per_sample];
	}

	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
	{
		ThrowException(m_adqInterfaces[adqDevListNum]->MultiRecordSetChannelMask(0XFF), "MultiRecordSetChannelMask");
		ThrowException(m_adqInterfaces[adqDevListNum]->MultiRecordSetup(nof_records, m_nof_samples), "MultiRecordSetup");
	}

	std::vector<bool> Master(m_adqInterfaces.size());

	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		Master[MasterIter->first] = true;
		for (int Slave = 0; Slave < MasterIter->second.size(); Slave++)
		{
			// This Sleep just to make it obvious whether the trigger stream starts with - or without - using the daisy chain.
			Sleep(100);
			ADQInterface* adqInterface = m_adqInterfaces[MasterIter->second[Slave]];
			ThrowException(adqInterface->DaisyChainEnable(1), "DaisyChainEnable");
			ThrowException(adqInterface->ArmTimestampSync(), "ArmTimestampSync");
			ThrowException(adqInterface->ArmTrigger(), "ArmTrigger");
			DaisyChainGetStatus(adqInterface, __LINE__);
		}
	}
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		unsigned int adqDevListNum = MasterIter->first;
		double sampleratehz;
		ADQInterface* adqInterface = m_adqInterfaces[adqDevListNum];
		if (!HardTrig)
		{
			ThrowException(adqInterface->GetSampleRate(0, &sampleratehz), "GetSampleRate");
			unsigned int internTrigPeriod = unsigned int(0.5 + TriggerPeriod * sampleratehz);
			ThrowException(adqInterface->SetInternalTriggerPeriod(internTrigPeriod), "SetInternalTriggerPeriod");
			uint32_t internTrigLowSamp = internTrigPeriod - internTrigHighSamp;
			ThrowException(adqInterface->SetInternalTriggerHighLow(internTrigHighSamp, internTrigLowSamp), "SetInternalTriggerHighLow");
		}
		ThrowException(adqInterface->DaisyChainEnable(1), "DaisyChainEnable");
//		ThrowException(adqInterface->ArmTimestampSync(), "ArmTimestampSync");
		ThrowException(adqInterface->ArmTrigger(), "ArmTrigger");
		DaisyChainGetStatus(adqInterface, __LINE__);
	}

	FILETIME StartTimeF;
	GetSystemTimeAsFileTime(&StartTimeF);
	ULARGE_INTEGER StartTime;
	StartTime.HighPart = StartTimeF.dwHighDateTime;
	StartTime.LowPart = StartTimeF.dwLowDateTime;

	//WaitForAcquisitionComplete(nof_records, StartTime);

	std::vector<unsigned int> nof_fetched_records(m_adqInterfaces.size());

	std::vector<ULARGE_INTEGER> LastElapsedTime(m_adqInterfaces.size());

	unsigned int min_nof_fetched_records = 0;
	std::vector<bool> DaisyChainStatusLogged(m_adqInterfaces.size());
	std::vector<bool> StreamOverflowLogged(m_adqInterfaces.size());
	int TrigSource = HardTrig ? 3 : 4;
	while (min_nof_fetched_records < nof_records)
	{
		unsigned int old_min_nof_fetched_records = min_nof_fetched_records;
		min_nof_fetched_records = UINT_MAX;
		bool GotAtLeastOne = false;
		for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
		{
			ADQInterface* adqInterface = m_adqInterfaces[adqDevListNum];
			// Frequently logs warnings?
			if (!StreamOverflowLogged[adqDevListNum])
			{
				int Err = adqInterface->GetStreamOverflow();
				if (Err)
				{
					printf("WARN : Stream overflow %d on %s\n", Err, adqInterface->GetBoardSerialNumber());
					StreamOverflowLogged[adqDevListNum] = true;
				}
			}
			// Sometimes logs warnings?
			if ((!DaisyChainStatusLogged[adqDevListNum]) && (DaisyChainGetStatus(adqInterface, __LINE__)))
				DaisyChainStatusLogged[adqDevListNum] = true;
			unsigned int nof_acquired_records = adqInterface->GetAcquiredRecords();
			if (nof_acquired_records > nof_fetched_records[adqDevListNum])
			{
				last_timestamp[adqDevListNum] = m_TimeStamp[adqDevListNum];
				m_TimeStamp[adqDevListNum] = 0;
				unsigned int bytes_per_sample;
				unsigned char ch_mask = 0xFF; // ch A+B+C+D+E+F+G+H
				unsigned long long thistimestamp;
				ThrowException(adqInterface->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
				ThrowException(adqInterface->GetDataWHTS(&(target_buffers[adqDevListNum][0]), NULL, &thistimestamp, m_nof_samples, bytes_per_sample,
					nof_fetched_records[adqDevListNum], 1, ch_mask, 0, m_nof_samples, ADQ_TRANSFER_MODE_NORMAL), "GetDataWHTS");
				m_TimeStamp[adqDevListNum] = thistimestamp;
				if (nof_fetched_records[adqDevListNum] == 0)
					first_timestamp[adqDevListNum] = thistimestamp;
				if (Master[adqDevListNum])
				{
					if (nof_fetched_records[adqDevListNum] > 0)
						// The timestamp for a internally-triggered (master) record does not get synchronised to 0.
						m_TimeStamp[adqDevListNum] -= first_timestamp[adqDevListNum];
					else
						m_TimeStamp[adqDevListNum] = 0;
				}
				unsigned long long tdiff = m_TimeStamp[adqDevListNum] - last_timestamp[adqDevListNum];
				double tdiff_d = (double)tdiff * 0.000000000025;
				unsigned long long ElapsedTime = GetElapsedTime(StartTime);
				double DiffElapsed = (ElapsedTime - LastElapsedTime[adqDevListNum].QuadPart) / 1E7; // 100ns intervals.
				LastElapsedTime[adqDevListNum].QuadPart = ElapsedTime;
				GotAtLeastOne = true;
				LastTrigTime = GetTickCount();
				Triggered = true;
				printf("Module %2d (%s) acquired: %4d records, Fetch record #%03d",
					adqDevListNum, adqInterface->GetBoardSerialNumber(), nof_acquired_records, nof_fetched_records[adqDevListNum]);
				printf(" elapsed time is %9llu (diff = %6.4f seconds), timestamp is %14llu (diff = %10.9f seconds)\n", ElapsedTime, DiffElapsed, m_TimeStamp[adqDevListNum], tdiff_d);
				nof_fetched_records[adqDevListNum]++;
			}
			min_nof_fetched_records = __min(min_nof_fetched_records, nof_fetched_records[adqDevListNum]);
		}
		if (min_nof_fetched_records > old_min_nof_fetched_records)
		{
			for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
			{
				if (Master[adqDevListNum])
				{
					unsigned long long DaisyTimeStamp = DaisyChainGetTriggerInformation2(adqDevListNum, old_min_nof_fetched_records);
				}
				for (unsigned int channel = 0; channel < target_buffers[adqDevListNum].size(); channel++)
					fwrite(target_buffers[adqDevListNum][channel], m_max_bytes_per_sample, m_nof_samples, DataFile);
			}
		}
		if (!GotAtLeastOne)
		{
			Sleep(1); // Wait some before polling again
			if (Triggered)
			{
				unsigned long WaitTime = GetTickCount() - LastTrigTime;
				if (WaitTime > 10000)
				{
					printf("WARN : Timeout on more triggers\n");
					break;
				}
			}
		}
	}
	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
	{
		/** Close MultiRecord **/
		SetTriggerIdleState(m_adqInterfaces[adqDevListNum]);
		m_adqInterfaces[adqDevListNum]->MultiRecordClose();
	}
	for (unsigned int adqDevListNum = 0; adqDevListNum < m_adqInterfaces.size(); adqDevListNum++)
		for (unsigned int channel = 0; channel < target_buffers[adqDevListNum].size(); channel++)
			delete[] target_buffers[adqDevListNum][channel];
	fclose(DataFile);
}

int main(int argc, const char* argv[])
{
	double TriggerPeriod = 0.5; // in sec
	bool HardTrig = false;
	std::vector<unsigned int> nof_records(1, 4);

	if (argc > 1)
		TriggerPeriod = atof(argv[1]);

	if (argc > 2)
	{
		HardTrig = (atoi(argv[2]) != 0);
	}

	if (argc > 3)
	{
		nof_records.resize(argc - 3);
		for (size_t Sequence = 0; Sequence < argc - 3; Sequence++)
			nof_records[Sequence] = atoi(argv[Sequence + 3]);
	}

	timeBeginPeriod(1);

	unsigned int adqDevList = 0;
	const unsigned int nof_samples = 600;
	const double Pre_Trig = -2.5e-07;
	unsigned int Group = 1;

	GetWHTSTest GetWHTSTest(nof_samples, Group);

	/** Set Up Collection with MultiRecord **/
	try {
		GetWHTSTest.Initialise();
		GetWHTSTest.detect_daisy_chain_order();
		GetWHTSTest.SetDaisyChain();
		std::vector<unsigned int> nof_channels(GetWHTSTest.adqInterfaces().size());
		for (unsigned int adqDevListNum = 0; adqDevListNum < GetWHTSTest.adqInterfaces().size(); adqDevListNum++)
			GetWHTSTest.SetInputRange(GetWHTSTest.adqInterfaces()[adqDevListNum]);
		for (size_t Sequence = 0; Sequence < nof_records.size(); Sequence++)
		{
			GetWHTSTest.SetTriggerMode(Pre_Trig, (Sequence == nof_records.size() - 1) ? HardTrig : false);
			printf("\nAcquiring: %000d records at %f.\n", nof_records[Sequence], TriggerPeriod);
			GetWHTSTest.AcquireSequence(nof_records[Sequence], TriggerPeriod, (Sequence == nof_records.size() - 1) ? HardTrig : false);
		}
	}
	catch(const char* ErrorSource)
	{
		printf("%s had an error\n", ErrorSource);
	}
	timeEndPeriod(1);
	return 0;
}
