#include <ADQAPI.h>

#include <vector>
#include <map>
#include <string>
#include <conio.h>

unsigned int g_max_nof_channels = 0;
unsigned int g_max_bytes_per_sample = 0;

void ThrowException(unsigned int status, const char* ErrorSource, bool ThrowIt = true) {
	if (status != 1)
	{
		if (ThrowIt)
			throw ErrorSource;
		else
			fprintf(stderr, "%s had an error\n", ErrorSource);
	}
}

void ThrowException(const char* ADQSerialNum, unsigned int status, const char* ErrorSource) {
	static std::map<std::string, bool> WasAnError;
	std::map<std::string, bool>::const_iterator WasAnErrorIter = WasAnError.find(ADQSerialNum);
	if (status != 1)
	{
		WasAnError[ADQSerialNum] = true;
		fprintf(stderr, "%s had an error on %s\r", ErrorSource, ADQSerialNum);
	}
	else if ((WasAnErrorIter != WasAnError.end()) && (WasAnError[ADQSerialNum]))
	{
		fprintf(stderr, "\n%s fault cleared on %s\n", ErrorSource, ADQSerialNum);
		WasAnError[ADQSerialNum] = false;
	}
}

unsigned long long GetElapsedTime(ULARGE_INTEGER const& StartTime)
{

	FILETIME NowTimeF;
	GetSystemTimeAsFileTime(&NowTimeF);
	ULARGE_INTEGER NowTime;
	NowTime.HighPart = NowTimeF.dwHighDateTime;
	NowTime.LowPart = NowTimeF.dwLowDateTime;

	return NowTime.QuadPart - StartTime.QuadPart;
}

void SetIdleState(ADQInterface* adqInterface, bool ThrowIt = true)
{
	int ADQType = adqInterface->GetADQType();
	if (ADQType == 8)
		ThrowException(adqInterface->SetInternalTriggerSyncMode(0), "SetInternalTriggerSyncMode", ThrowIt);
	ThrowException(adqInterface->DisarmTimestampSync(), "DisarmTimestampSync", ThrowIt);
	ThrowException(adqInterface->DisarmTrigger(), "DisarmTrigger", ThrowIt);
	if (ADQType == 8)
	{
		ThrowException(adqInterface->DaisyChainEnable(0), "DaisyChainEnable", ThrowIt);
		ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset", ThrowIt);
	}
}

void Initialise(void* adqCtrlUnit, int adqDevListNum,
	int nof_records, int nof_samples, int clock_source, std::vector<ADQInterface*>& adqInterfaces)
{
	// Opens communication channel to a certain ADQ device
	ThrowException(ADQControlUnit_OpenDeviceInterface(adqCtrlUnit, adqDevListNum), "ADQControlUnit_OpenDeviceInterface");

	// Make this device ready to use
	ThrowException(ADQControlUnit_SetupDevice(adqCtrlUnit, adqDevListNum), "ADQControlUnit_SetupDevice");

	ADQInterface* adqInterface = ADQControlUnit_GetADQ(adqCtrlUnit, adqDevListNum + 1);
	adqInterfaces[adqDevListNum] = adqInterface;
	ThrowException(adqInterface ? 1 : 0, "ADQControlUnit_GetADQ");

	ThrowException(IS_VALID_DLL(adqInterfaces[adqDevListNum]), "IS_VALID_DLL");
	//	ThrowException(adqInterface->ResetDevice(2), "ResetDevice");

	int* FWRevision = adqInterfaces[adqDevListNum]->GetRevision();
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	printf("Serial number %s FW Revision %d\n", SerialNumber.c_str(), *FWRevision);

	ThrowException(adqInterface->MultiRecordSetup(nof_records ? nof_records : 1, nof_samples), "MultiRecordSetup");
	/** Allocate memory for storing data **/

	g_max_nof_channels = __max(g_max_nof_channels, adqInterfaces[adqDevListNum]->GetNofChannels());

	unsigned int bytes_per_sample;
	ThrowException(adqInterface->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
	g_max_bytes_per_sample = __max(g_max_bytes_per_sample, bytes_per_sample);

	SetIdleState(adqInterface);

	// 3 - Internal clock source, external 10 MHz reference from PXIsync
	// 1 - Internal clock source
	ThrowException(adqInterfaces[adqDevListNum]->GetBoardSerialNumber(), adqInterfaces[adqDevListNum]->SetClockSource(clock_source), "SetClockSource");
}

void SetTriggerMode(ADQInterface* adqInterface, double TriggerPeriod)
{
	std::string SerialNumber = adqInterface->GetBoardSerialNumber();
	//	How does one enable internal trigger synchronisation for ADQ8, if this is not supported?
	double sampleratehz;
	ThrowException(adqInterface->GetSampleRate(0, &sampleratehz), "GetSampleRate");
	unsigned int internTrigPeriod = unsigned int(0.5 + TriggerPeriod * sampleratehz);
	ThrowException(adqInterface->SetTriggerMode(4), "SetTriggerMode"); // Internal trigger mode
	ThrowException(adqInterface->SetInternalTriggerPeriod(internTrigPeriod), "SetInternalTriggerPeriod");
	ThrowException(adqInterface->SetupTimestampSync(0, 4), "SetupTimestampSync");
	//	ThrowException(adqInterface->ArmTimestampSync(), "ArmTimestampSync");
}


int main(int argc, const char* argv[])
{
	void* adqCtrlUnit = NULL;
	double TriggerPeriod = 0.5; // in sec
	unsigned int nof_records = 4;
	int clock_source = 3;

	if (argc > 1)
		TriggerPeriod = atof(argv[1]);

	if (argc > 2)
		nof_records = atoi(argv[2]);

	if (argc > 3)
		clock_source = atoi(argv[3]);

	std::vector<ADQInterface*> adqInterfaces;

	unsigned int adqDevList = 0;
	unsigned int nof_samples = 500;
	void** target_buffers = NULL;
	std::vector<unsigned int> nof_channels;

	/** Set Up Collection with MultiRecord **/
	try {
		void* adqCtrlUnit = CreateADQControlUnit();   // Creates an ADQControlUnit called adq_cu
		ThrowException(adqCtrlUnit != NULL, "CreateADQControlUnit");

		ThrowException(ADQControlUnit_EnableErrorTrace(adqCtrlUnit, LOG_LEVEL_INFO, "."), "ADQControlUnit_EnableErrorTrace");

		struct ADQInfoListEntry* adqInfoStruct;

		ThrowException(ADQControlUnit_ListDevices(adqCtrlUnit, &adqInfoStruct, &adqDevList), "ADQControlUnit_ListDevices");

		adqInterfaces.resize(adqDevList);
		nof_channels.resize(adqDevList);

		int APIRevision = ADQAPI_GetRevision();
		printf("API Revision %d\n", APIRevision);

		std::vector<unsigned long long> timestamp(adqDevList),
			last_timestamp(adqDevList);

		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
		{
			Initialise(adqCtrlUnit, adqDevListNum,
				nof_records, nof_samples, clock_source, adqInterfaces);
		}
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
		{
			SetTriggerMode(adqInterfaces[adqDevListNum], TriggerPeriod);
		}

		target_buffers = static_cast<void**>(malloc(g_max_nof_channels * sizeof(void*)));
		for (unsigned int channel = 0; channel < g_max_nof_channels; channel++)
		{
			target_buffers[channel] = malloc(nof_samples*g_max_bytes_per_sample);
			ThrowException(target_buffers[channel] ? 1 : 0, "malloc"); // Flag failure if memory allocation failed
		}

		printf("\nStarting collection\n");
		fflush(stdout);

		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
		{
			// This Sleep just to make it obvious whether the trigger stream starts with - or without - using the daisy chain.
			if (adqDevListNum > 0)
				Sleep(100);
			ThrowException(adqInterfaces[adqDevListNum]->ArmTrigger(), "ArmTrigger");
			int ADQType = adqInterfaces[adqDevListNum]->GetADQType();
			if (ADQType == 8)
				ThrowException(adqInterfaces[adqDevListNum]->DaisyChainEnable(1), "DaisyChainEnable");
		}

		unsigned int nof_fetched_records = 0;
		unsigned char ch_mask = 0xFF; // ch A+B+C+D+E+F+G+H

		FILETIME StartTimeF;
		GetSystemTimeAsFileTime(&StartTimeF);
		ULARGE_INTEGER StartTime;
		StartTime.HighPart = StartTimeF.dwHighDateTime;
		StartTime.LowPart = StartTimeF.dwLowDateTime;
		std::vector<ULARGE_INTEGER> LastElapsedTime(adqDevList);

		printf("Acquiring: %000d records at %f.\n", nof_records, TriggerPeriod);
		std::vector<unsigned int> nof_acquired_records(adqDevList);
		while ((nof_fetched_records < nof_records) || (nof_records == 0))
		{
			if (_kbhit())
				break;
			unsigned int min_nof_acquired_records = UINT_MAX;
			for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
			{
				if (_kbhit())
					break;
				ThrowException(adqInterfaces[adqDevListNum]->GetBoardSerialNumber(), adqInterfaces[adqDevListNum]->SetClockSource(clock_source), "SetClockSource");
				nof_acquired_records[adqDevListNum] = adqInterfaces[adqDevListNum]->GetAcquiredRecords();
				min_nof_acquired_records = __min(min_nof_acquired_records, nof_acquired_records[adqDevListNum]);
			}
			if (min_nof_acquired_records > nof_fetched_records)
			{
				for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
				{
					printf("Module %2d acquired: %4d records, Fetch record #%0d",
						adqDevListNum, nof_acquired_records[adqDevListNum], nof_fetched_records);
					/** Transfer Data from ADQ to host - records one by one**/
					/* Re-use target buffer allocated on each call */
					last_timestamp[adqDevListNum] = timestamp[adqDevListNum];
					timestamp[adqDevListNum] = 0;
					unsigned int bytes_per_sample;
					ThrowException(adqInterfaces[adqDevListNum]->GetNofBytesPerSample(&bytes_per_sample), "GetNofBytesPerSample");
					ThrowException(adqInterfaces[adqDevListNum]->GetDataWHTS(target_buffers, NULL, &timestamp[adqDevListNum], nof_samples, bytes_per_sample,
						nof_fetched_records, 1, ch_mask, 0, nof_samples, ADQ_TRANSFER_MODE_NORMAL), "GetDataWHTS");
					//					if (Master[adqDevListNum])
					//						ThrowException(adqInterfaces[adqDevListNum]->DaisyChainGetTriggerInformation(4, )
					unsigned long long tdiff = timestamp[adqDevListNum] - last_timestamp[adqDevListNum];
					double tdiff_d = (double)tdiff * 0.000000000025;
					unsigned long long ElapsedTime = GetElapsedTime(StartTime);
					double DiffElapsed = (ElapsedTime - LastElapsedTime[adqDevListNum].QuadPart) / 1E7; // 100ns intervals.
					printf(" elapsed time is %8llu (diff = %5.4f seconds), timestamp is %14llu (diff = %5.4f seconds)\n", ElapsedTime, DiffElapsed, timestamp[adqDevListNum], tdiff_d);
					LastElapsedTime[adqDevListNum].QuadPart = ElapsedTime;
				}
				nof_fetched_records++;
			}
			else
			{
				if (nof_records == 0)
				{
					nof_fetched_records = 0;
					for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
					{
						ThrowException(adqInterfaces[adqDevListNum]->MultiRecordClose(), "MultiRecordClose");
						ThrowException(adqInterfaces[adqDevListNum]->MultiRecordSetup(1, nof_samples), "MultiRecordSetup");
						ThrowException(adqInterfaces[adqDevListNum]->ArmTrigger(), "ArmTrigger");
					}
				}
				Sleep(1); // Wait some before polling again
			}
		}
		for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
		{
			/** Close MultiRecord **/
			SetIdleState(adqInterfaces[adqDevListNum], false);
			adqInterfaces[adqDevListNum]->MultiRecordClose();
		}
	}
	catch (const char* ErrorSource)
	{
		fprintf(stderr, "%s had an error\n", ErrorSource);
	}
	// should one call ADQControlUnit_CloseDevice? I get an access violation.
//	for (unsigned int adqDevListNum = 0; adqDevListNum < adqDevList; adqDevListNum++)
//		(adqInterfaces[adqDevListNum], adqDevListNum);
	if (adqCtrlUnit)
		DeleteADQControlUnit(adqCtrlUnit);
	return 0;
}
