#include "utils.h"

#include <ADQAPI.h>
#include <stdio.h>
#include <conio.h>

static const int num_channels = 8;
static const unsigned int MaxDaisyJitter = 100;

CUtils::SChannelSettings::SChannelSettings() 
{
	m_input_range = 1000.0;		//mV
	dc_offset = 0;
	m_input_impedance = false;
}

CUtils::SModuleSettings::SModuleSettings()
{
	m_level_trig_channel = 7;	// Channel H
	m_level_trig_level = 1024;
	m_level_trig_arm_hyst = 1;
	m_level_trig_edge = 1;
	m_trigger_source = 23;		// daisy trigger
	m_pretrig_samples = 0;
	m_trigger_delay_samples = 248;
	m_channelsettings.resize(num_channels);
}

CUtils::SSettings::SSettings(int clock_source)
{
	m_clock_source = clock_source;
	m_ref_clock_delay_ps = 0;
	m_sample_skip_factor = 1;
	m_number_of_records = 100;
	m_record_length = 600;
}

void CUtils::list_devices()
{
	m_adq_cu = CreateADQControlUnit();
	ThrowException(m_adq_cu != NULL, "CreateADQControlUnit");

	ThrowException(ADQControlUnit_EnableErrorTrace(m_adq_cu, LOG_LEVEL_RTONLY, "."), "ADQControlUnit_EnableErrorTrace");

	printf("Enumerating devices..\n");
	m_nof_adqs = ADQControlUnit_FindDevices(m_adq_cu);

	m_nof_adqs = ADQControlUnit_NofADQ(m_adq_cu);

	int nof_failed_devices = ADQControlUnit_GetFailedDeviceCount(m_adq_cu);
	if (nof_failed_devices > 0)
		ThrowException(0, "ADQControlUnit_GetFailedDeviceCount");

	if (m_nof_adqs == 0)
		ThrowException(0, "No devices found");

	m_adqInterfaces.resize(m_nof_adqs);

	for (int adq_num = 0; adq_num < m_nof_adqs; adq_num++)
	{
		ThrowException(ADQControlUnit_OpenDeviceInterface(m_adq_cu, adq_num), "ADQControlUnit_OpenDeviceInterface");
		m_adqInterfaces[adq_num] = ADQControlUnit_GetADQ(m_adq_cu, adq_num + 1);
		ThrowException(m_adqInterfaces[adq_num] ? 1 : 0, "ADQControlUnit_GetADQ");
		std::string SerialNumber = m_adqInterfaces[adq_num]->GetBoardSerialNumber();
	}
}

CUtils::CUtils(int clock_source) :
	m_settings(clock_source)
{
}

CUtils::~CUtils()
{
	close_all_devices();
	for (int adq_num = 0; adq_num < m_nof_adqs; adq_num++)
		ThrowException(ADQControlUnit_CloseDevice(m_adq_cu, adq_num), "ADQControlUnit_CloseDevice", false);
	for (size_t adq_num = 0; adq_num < m_target_buffers.size(); adq_num++)
		for (size_t records_completed = 0; records_completed < m_target_buffers[adq_num].size(); records_completed++)
			for (size_t ch = 0; ch < m_target_buffers[adq_num][records_completed].size(); ch++)
				delete[] m_target_buffers[adq_num][records_completed][ch];
	DeleteADQControlUnit(m_adq_cu);
}

void CUtils::detect_daisy_chain_order()
{
	// First disable all outputs
	for (int adq_num = 0; adq_num < m_nof_adqs; adq_num++)
	{
		ThrowException(m_adqInterfaces[adq_num]->DaisyChainEnableOutput(0), "DaisyChainEnableOutput");
		ThrowException(m_adqInterfaces[adq_num]->DaisyChainSetOutputState(0), "DaisyChainSetOutputState");
	}
	std::vector<int> next_device(m_nof_adqs, -1);
	std::vector<int> prev_device(m_nof_adqs, -1);
	for (int adq_out_num = 0; adq_out_num < m_nof_adqs; adq_out_num++)
	{
		// Determine which digitizer is not pointed to by any device.
		// Filter out the last slave(next_device == -1)
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainEnableOutput(1), "DaisyChainEnableOutput");
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainSetOutputState(1), "DaisyChainSetOutputState");
		for (int adq_in_num = 0; adq_in_num < m_nof_adqs; adq_in_num++)
		{
			if (adq_in_num == adq_out_num)
				continue;
			unsigned int InputState = 0;
			ThrowException(m_adqInterfaces[adq_in_num]->DaisyChainGetInputState(&InputState), "DaisyChainSetOutputState");
			if (InputState == 1)
			{
				if (next_device[adq_out_num] != -1)
					ThrowException(0, "Device input already connected");
				next_device[adq_out_num] = adq_in_num;
				if (prev_device[adq_in_num] != -1)
					ThrowException(0, "Device output already connected");
				prev_device[adq_in_num] = adq_out_num;
			}
		}
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainSetOutputState(0), "DaisyChainSetOutputState");
		ThrowException(m_adqInterfaces[adq_out_num]->DaisyChainEnableOutput(0), "DaisyChainEnableOutput");
	}
	m_device_position.clear();
	int nof_unconnected_slaves = 0;
	for (int adq_num = 0; adq_num < m_nof_adqs; adq_num++)
	{
		if (next_device[adq_num] == -1)
			nof_unconnected_slaves++;
		if (prev_device[adq_num] == -1)
			m_device_position[adq_num] = std::vector<int>();
	}
	if (nof_unconnected_slaves != m_device_position.size())
		ThrowException(0, "At least one incomplete chain");

	// Walk through the next_device list to determine the devices' positions in
	// the daisy chain.
	for (std::map<int, std::vector<int>>::iterator DaisyIter = m_device_position.begin();
		DaisyIter != m_device_position.end(); DaisyIter++)
	{
		int next_slave_num = DaisyIter->first;
		printf("Master %s is %d\n", m_adqInterfaces[next_slave_num]->GetBoardSerialNumber(), next_slave_num);
		m_settings.m_masterModuleSettings.insert(std::pair<int, SModuleSettings>(DaisyIter->first, SModuleSettings()));
		while (true)
		{
			next_slave_num = next_device[next_slave_num];
			if (next_slave_num == -1)
				break;
			printf("Slave %s is %d\n", m_adqInterfaces[next_slave_num]->GetBoardSerialNumber(), next_slave_num);
			DaisyIter->second.push_back(next_slave_num);
			m_settings.m_slaveModuleSettings.insert(std::pair<int, SModuleSettings>(next_slave_num, SModuleSettings()));
		}
	}
}

void CUtils::fill_daisy_chain_device_information()
{
}

void CUtils::setup_collection(int adq_num, SModuleSettings const& modulesettings) const
{
	ADQInterface* adqInterface = m_adqInterfaces[adq_num];
	std::vector<unsigned int> m_enabled_channels;

	std::string SerialNumber = adqInterface->GetBoardSerialNumber();

	bool TerminatingModule = false;
	if (SerialNumber == "SPD-06916")
		// This is the terminating module of the connected trigger and input cabling transmission line.
		// THIS WILL NEED TO BE CHANGED for other hardware setups.
		TerminatingModule = true;

	for (unsigned int ch = 1; ch <= modulesettings.m_channelsettings.size(); ch++)
	{
		float result;
		SChannelSettings const& ChannelSettings = modulesettings.m_channelsettings[ch - 1];
		if (ChannelSettings.m_input_range == 0)
			continue;
		ThrowException(adqInterface->SetInputRange(ch, ChannelSettings.m_input_range, &result), "SetInputRange");
		ThrowException(adqInterface->SetAdjustableBias(ch, ChannelSettings.dc_offset), "SetAdjustableBias");
		if ((TerminatingModule) && (ch == 2))
			// THIS WILL NEED TO BE CHANGED for other hardware setups.
			ThrowException(adqInterface->SetInputImpedance(ch, 0), "SetInputImpedance");
		else
			ThrowException(adqInterface->SetInputImpedance(ch, ChannelSettings.m_input_impedance ? 0 : 1), "SetInputImpedance");
	}

	ThrowException(adqInterface->AdjustClockReferenceDelay(m_settings.m_ref_clock_delay_ps), "AdjustClockReferenceDelay");

	ThrowException(adqInterface->SetClockSource(m_settings.m_clock_source), "SetClockSource " + SerialNumber, false);
	ThrowException(adqInterface->EnableClockRefOut(1), "EnableClockRefOut");

	if (modulesettings.m_trigger_source == 3)
	{
		unsigned int level_trigger_channel_mask = 1 << modulesettings.m_level_trig_channel;
		int level_trig_level = modulesettings.m_level_trig_level;
		int level_trig_edge = modulesettings.m_level_trig_edge;
		int level_trig_arm_hyst = modulesettings.m_level_trig_arm_hyst;

		ThrowException(m_adqInterfaces[adq_num]->SetupLevelTrigger(
			&level_trig_level,
			&level_trig_edge,
			&level_trig_arm_hyst,
			level_trigger_channel_mask, 0), "SetupLevelTrigger");
	}

	ThrowException(adqInterface->SetTriggerMode(modulesettings.m_trigger_source), "SetTriggerMode");
	ThrowException(adqInterface->SetPreTrigSamples(modulesettings.m_pretrig_samples), "SetPreTrigSamples");
	ThrowException(adqInterface->SetSampleSkip(m_settings.m_sample_skip_factor), "SetSampleSkip");
	ThrowException(adqInterface->SetTriggerHoldOffSamples(modulesettings.m_trigger_delay_samples), "SetTriggerHoldOffSamples");
	ThrowException(adqInterface->MultiRecordSetChannelMask(0XFF), "MultiRecordSetChannelMask");
	ThrowException(adqInterface->MultiRecordSetup(m_settings.m_number_of_records, m_settings.m_record_length), "MultiRecordSetup");
}

void CUtils::setup_collection() const
{
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		std::map<int, SModuleSettings>::const_iterator MasterModIter = m_settings.m_masterModuleSettings.find(MasterIter->first);
		setup_collection(MasterIter->first, MasterModIter->second);
		for (size_t position = 0; position < MasterIter->second.size(); position++)
		{
			std::map<int, SModuleSettings>::const_iterator SlaveModIter = m_settings.m_slaveModuleSettings.find(MasterIter->second[position]);
			setup_collection(MasterIter->second[position], SlaveModIter->second);
		}
	}
}

void CUtils::setup_daisy_chain(int adq_num, int position, SModuleSettings const& modulesettings) const
{
	ADQInterface* adqInterface = m_adqInterfaces[adq_num];
	ThrowException(adqInterface->DaisyChainEnable(0), "DaisyChainEnable");
	ThrowException(adqInterface->DaisyChainReset(), "DaisyChainReset");
	ThrowException(adqInterface->DaisyChainSetMode((position == 0) ? 1 : 0), "DaisyChainSetMode");
	if (position == 0)
		ThrowException(adqInterface->DaisyChainSetupLevelTrigger(
			modulesettings.m_level_trig_channel+1,
			modulesettings.m_level_trig_level,
			modulesettings.m_level_trig_arm_hyst,
			modulesettings.m_level_trig_edge), "DaisyChainSetupLevelTrigger");
	ThrowException(adqInterface->DaisyChainSetupOutput(1, 0, 10 * 4), "DaisyChainSetupOutput");
	ThrowException(adqInterface->DaisyChainEnableOutput(1), "DaisyChainEnableOutput");
	if (position == 0)
		ThrowException(adqInterface->DaisyChainSetTriggerSource(3), "DaisyChainSetTriggerSource");
}

void CUtils::setup_daisy_chain() const
{
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		std::map<int, SModuleSettings>::const_iterator MasterModIter = m_settings.m_masterModuleSettings.find(MasterIter->first);
		setup_daisy_chain(MasterIter->first, 0, MasterModIter->second);
		for (int position = 0; position < MasterIter->second.size(); position++)
		{
			std::map<int, SModuleSettings>::const_iterator SlaveModIter = m_settings.m_slaveModuleSettings.find(MasterIter->second[position]);
			setup_daisy_chain(MasterIter->second[position], position + 1, SlaveModIter->second);
		}
	}
}

void CUtils::adjust_pre_trigger_samples(int adq_num, int position, SModuleSettings& modulesettings)
{
	unsigned int additional_pretrigger_samples;
	ThrowException(m_adqInterfaces[adq_num]->DaisyChainGetNofPretriggerSamples(position, m_settings.m_sample_skip_factor,
		&additional_pretrigger_samples), "DaisyChainGetNofPretriggerSamples");
	modulesettings.m_pretrig_samples += additional_pretrigger_samples;
//  Required if DaisyChainGetTriggerInformation uses trigger source 3.
//	modulesettings.m_pretrig_samples += 4;
}

void CUtils::adjust_pre_trigger_samples()
{
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		std::map<int, SModuleSettings>::iterator MasterModIter = m_settings.m_masterModuleSettings.find(MasterIter->first);
		adjust_pre_trigger_samples(MasterIter->first, 0, MasterModIter->second);
		for (int position = 0; position < MasterIter->second.size(); position++)
		{
			std::map<int, SModuleSettings>::iterator SlaveModIter = m_settings.m_slaveModuleSettings.find(MasterIter->second[position]);
			adjust_pre_trigger_samples(MasterIter->second[position], position + 1, SlaveModIter->second);
		}
	}
}

void CUtils::arm_all() const
{
	for (size_t adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
		m_adqInterfaces[adq_num]->DisarmTrigger();
	for (size_t adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
	{
		m_adqInterfaces[adq_num]->ArmTrigger();
	}
}

void CUtils::enable_daisy_chain_all() const
{
	// Enable slaves
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		for (size_t position = 0; position < MasterIter->second.size(); position++)
		{
			int adq_num = MasterIter->second[position];
			m_adqInterfaces[adq_num]->DaisyChainEnable(1);
		}
	}
	// Enable masters
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		int adq_num = MasterIter->first;
		m_adqInterfaces[adq_num]->DaisyChainEnable(1);
	}
}

void CUtils::send_triggers(int adq_num, SModuleSettings const& modulesettings) const
{
	if (modulesettings.m_trigger_source != 1)
		return;
	for (unsigned int record = 0; record < m_settings.m_number_of_records; record++)
	{
		ThrowException(m_adqInterfaces[adq_num]->SWTrig(), "SWTrig");
	}
}

void CUtils::send_triggers() const
{
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		std::map<int, SModuleSettings>::const_iterator MasterModIter = m_settings.m_masterModuleSettings.find(MasterIter->first);
		send_triggers(MasterIter->first, MasterModIter->second);
		for (size_t position = 0; position < MasterIter->second.size(); position++)
		{
			std::map<int, SModuleSettings>::const_iterator SlaveModIter = m_settings.m_slaveModuleSettings.find(MasterIter->second[position]);
			send_triggers(MasterIter->second[position], SlaveModIter->second);
		}
	}
}

bool CUtils::wait_for_acquired_all()
{
	unsigned int min_nof_acquired_records = 0;
	DWORD LastTrigTime = GetTickCount();
	printf("Waiting for data..\n"); 
	fflush(stdout);
	std::vector<unsigned int> nof_acquired_records(m_adqInterfaces.size());
	while (min_nof_acquired_records < m_settings.m_number_of_records)
	{
		if (_kbhit())
			return false;
		min_nof_acquired_records = UINT_MAX;
		bool GotAtLeastOne = false;
		for (unsigned int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
		{
			ADQInterface* adqInterface = m_adqInterfaces[adq_num];
			unsigned int new_nof_acquired_records = adqInterface->GetAcquiredRecords();
			if (new_nof_acquired_records > nof_acquired_records[adq_num])
			{
				GotAtLeastOne = true;
				nof_acquired_records[adq_num] = new_nof_acquired_records;
				LastTrigTime = GetTickCount();
			}
			min_nof_acquired_records = __min(min_nof_acquired_records, new_nof_acquired_records);
		}
		if (GotAtLeastOne)
		{
			printf("%d triggers detected\r", min_nof_acquired_records);
		}
		else
		{
			Sleep(1);
			if ((min_nof_acquired_records > 0) && (GetTickCount() - LastTrigTime > 2000))
			{
				m_settings.m_number_of_records = min_nof_acquired_records;
				printf("WARNING: timed out on more triggers\n");
				break;
			}
		}
	}
	printf("All data collected\n");
	return true;
}

void CUtils::collect_data(int adq_num)
{
	bool StreamOverflowLogged=false;
	m_target_buffers[adq_num].resize(m_settings.m_number_of_records);

	for (unsigned int records_completed = 0; records_completed < m_settings.m_number_of_records; records_completed++)
	{
		m_target_buffers[adq_num][records_completed].resize(num_channels);
		for (int ch = 0; ch < num_channels; ch++)
			m_target_buffers[adq_num][records_completed][ch] = new unsigned short[m_settings.m_record_length];
		ADQInterface* adqInterface = m_adqInterfaces[adq_num];
		int Err = adqInterface->GetStreamOverflow();
		if ((Err) && (!StreamOverflowLogged))
		{
			printf("WARN : Stream overflow %d on %s\n", Err, adqInterface->GetBoardSerialNumber());
			StreamOverflowLogged = true;
		}

		ThrowException(adqInterface->GetDataWHTS(&(static_cast<void*>(m_target_buffers[adq_num][records_completed][0])), NULL, NULL, m_settings.m_record_length, sizeof(short),
			records_completed, 1, 0XFF, 0, m_settings.m_record_length, ADQ_TRANSFER_MODE_NORMAL), "GetDataWHTS");
	}
}
void CUtils::collect_data_all()
{
	m_timeStamps.resize(m_settings.m_number_of_records);
	std::vector <ADQDaisyChainTriggerInformation> trig_info(m_settings.m_number_of_records);
	const double PicoSec25 = 25E-12;
	int MasterCount = 0;
	for (std::map<int, std::vector<int>>::const_iterator MasterIter = m_device_position.begin();
		MasterIter != m_device_position.end(); MasterIter++)
	{
		if (MasterIter->second.size() == 0)
			// Ignored modules.
			continue;
		MasterCount++;
		SModuleSettings const& masterModuleSettings = m_settings.m_masterModuleSettings[MasterIter->first];
		std::vector<ADQDaisyChainDeviceInformation> device_info_array(MasterIter->second.size() + 1);
		std::vector<std::vector<double>> ExtendedPrecision(trig_info.size());
		std::vector<std::vector<int64_t>> RecordStart(trig_info.size());
		device_info_array[0].Position = 0;
		device_info_array[0].PretriggerSamples = masterModuleSettings.m_pretrig_samples;
		device_info_array[0].TriggerDelaySamples = masterModuleSettings.m_trigger_delay_samples;
		RecordStart.resize(trig_info.size());
		for (size_t record = 0; record < trig_info.size(); record++)
		{
			RecordStart[record].resize(device_info_array.size());
			ExtendedPrecision[record].resize(device_info_array.size());
			trig_info[record].ExtendedPrecision = &(ExtendedPrecision[record][0]);
			trig_info[record].RecordStart = &(RecordStart[record][0]);
		}
		for (unsigned int position = 1; position <= MasterIter->second.size(); position++)
		{
			SModuleSettings const& slaveModuleSettings = m_settings.m_slaveModuleSettings[MasterIter->second[position - 1]];
			device_info_array[position].Position = position;
			device_info_array[position].PretriggerSamples = slaveModuleSettings.m_pretrig_samples;
			device_info_array[position].TriggerDelaySamples = slaveModuleSettings.m_trigger_delay_samples;
		}

		ThrowException(m_adqInterfaces[MasterIter->first]->DaisyChainGetTriggerInformation(
			masterModuleSettings.m_trigger_source,
			masterModuleSettings.m_level_trig_edge,
			masterModuleSettings.m_level_trig_level,
			masterModuleSettings.m_level_trig_channel + 1,
			m_settings.m_number_of_records,
			m_settings.m_record_length,
			m_settings.m_sample_skip_factor,
			&(device_info_array[0]),
			unsigned int(device_info_array.size()),
			&(trig_info[0])), "DaisyChainGetTriggerInformation");

		m_RecordStart.resize(trig_info.size());
		for (size_t record = 0; record < trig_info.size(); record++)
		{
			uint64_t MasterTimeStamp = trig_info[record].Timestamp;
			m_RecordStart[record].resize(m_adqInterfaces.size());
			m_RecordStart[record][MasterIter->first] = RecordStart[record][0];
			for (unsigned int position = 1; position <= MasterIter->second.size(); position++)
				m_RecordStart[record][MasterIter->second[position-1]] = RecordStart[record][position];
			m_timeStamps[record] += (trig_info[record].Timestamp * PicoSec25);
		}
	}
	for (size_t record = 0; record < trig_info.size(); record++)
		m_timeStamps[record] /= MasterCount;
	m_target_buffers.resize(m_adqInterfaces.size());
	for (unsigned int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
		collect_data(adq_num);
}

void CUtils::save_data() const
{
	FILE* DataFile = fopen("DaisyDemo.hioc", "wt");
	if (!DataFile)
		ThrowException(0, "Unable to open data file");
	fprintf(DataFile, "Pre_Trig 0\n");
	fprintf(DataFile, "Num_Segments %d\n", m_settings.m_number_of_records);
	fprintf(DataFile, "Num_Background 0\n");
	fprintf(DataFile, "Trig_Mode 3\n");
	fprintf(DataFile, "Points_Per_Segment %d\n", m_settings.m_record_length-MaxDaisyJitter);

	char SpectrometerCount = 'A';
	for (unsigned int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
	{
		std::map<int, SModuleSettings>::const_iterator ModIter = m_settings.m_masterModuleSettings.find(adq_num);
		if (ModIter == m_settings.m_masterModuleSettings.end())
		{
			ModIter = m_settings.m_slaveModuleSettings.find(adq_num);
			if (ModIter == m_settings.m_slaveModuleSettings.end())
				ThrowException(0, "m_settings.m_slaveModuleSettings.find");
		}
		static const int SpectrometersPerMod = 2;
		static const int ChansPerSpectrometer = 4;
		for (unsigned int SpectrometerInMod = 0; SpectrometerInMod < SpectrometersPerMod; SpectrometerInMod++)
		{
			for (unsigned int ch = 0; ch < num_channels / SpectrometersPerMod; ch++)
			{
				fprintf(DataFile, "Spectr-%c-Range_Chan%d %f\n", SpectrometerCount, 1 + ch % ChansPerSpectrometer,
					ModIter->second.m_channelsettings[ch + ChansPerSpectrometer * SpectrometerInMod].m_input_range / 1000);
				fprintf(DataFile, "Spectr-%c-Offset_Chan%d %f\n", SpectrometerCount, 1 + ch % ChansPerSpectrometer, 0.0);
			}
			SpectrometerCount++;
		}
	}
	fprintf(DataFile, "TimeVec ");
	for (unsigned int records_completed = 0; records_completed < m_settings.m_number_of_records; records_completed++)
	{
		// With respect to first timestamp.
		fprintf(DataFile, "%f", m_timeStamps[records_completed] - m_timeStamps[0]);
		if (records_completed < m_settings.m_number_of_records - 1)
			fprintf(DataFile, ",");
		else
			fprintf(DataFile, "\n");
	}
	fclose(DataFile);
	DataFile = fopen("DaisyDemo.hioc", "ab");
	if (!DataFile)
		ThrowException(0, "Unable to open data file");
	// Teminate text with an EOF charachter, prior to binary data storage.
	fputc(char(26), DataFile);

	for (unsigned int records_completed = 0; records_completed < m_settings.m_number_of_records; records_completed++)
	{
		for (unsigned int adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
		{
			for (unsigned int ch = 0; ch < m_target_buffers[adq_num][records_completed].size(); ch++)
			{
				int64_t DaisyJitter = m_RecordStart[records_completed][adq_num];
				if (DaisyJitter < 0)
					fwrite(m_target_buffers[adq_num][records_completed][ch] - (DaisyJitter / 40), sizeof(short), m_settings.m_record_length - MaxDaisyJitter, DataFile);
				else
				{
					std::vector<short> Null(DaisyJitter / 40);
					fwrite(&(Null[0]), sizeof(short), Null.size(), DataFile);
					fwrite(m_target_buffers[adq_num][records_completed][ch], sizeof(short), m_settings.m_record_length - Null.size(), DataFile);
				}
			}
		}
	}

	fclose(DataFile);
}

//////////////////////////////////////////////////////////////////////////////////////////
//																						//
//	void CUtils::ThrowException(unsigned int status, const char* ErrorSource,			//
//		bool ThrowIt /*= true*/)														//
//																						//
//	Description:																		//
//		Throws ErrorSource if the status is bad.										//
//		Optionally, logs a warning instead, if ThrowIt is false.						//
//																						//
//////////////////////////////////////////////////////////////////////////////////////////
void CUtils::ThrowException(unsigned int status, std::string const& ErrorSource, bool ThrowIt /*= true*/) const
{
	if (status != 1)
	{
		if (ThrowIt)
			throw ErrorSource;
		else
			printf("WARN : %s had an error\n", ErrorSource.c_str());
	}
}

void CUtils::close_all_devices() const
{
	for (size_t adq_num = 0; adq_num < m_adqInterfaces.size(); adq_num++)
	{
		ThrowException(m_adqInterfaces[adq_num]->DisarmTrigger(), "DisarmTrigger", false);
		ThrowException(m_adqInterfaces[adq_num]->MultiRecordClose(), "MultiRecordClose", false);
	}
}