#!/usr/bin/env python3
# vim:fileencoding=utf-8
"""
 Copyright 2018 Teledyne Signal Processing Devices Sweden AB
"""

import os
from copy import deepcopy

from utils import (
    list_devices,
    detect_daisy_chain_order,
    NoDeviceFound,
    setup_collection,
    setup_daisy_chain,
    setup_sdcard,
    arm_all,
    enable_daisy_chain_all,
    get_daisy_chain_nof_pretrigger_samples,
    fill_daisy_chain_device_information,
    close_all_devices,
    send_triggers,
    collect_data_all,
    read_delays,
    update_delays,
    plot_daisy_chain,
    save_data,
    wait_for_acquired_all,
)


def main():
    enabled_channels = [1, 2]

    daisy_chain_settings = {
        "position": 0,
        "level_trig_channel": 1,
        "level_trig_edge": 1,  # 1: Rising, 0: Falling
        "level_trig_level": 0,
        "level_trig_arm_hyst": 500,
        "trigger_source": 3,  # 1: Software trigger, 3: Level trigger
        "compare_trig_info_sdcard": False,
    }

    settings = {
        "data_path": os.path.join("data", os.path.splitext(__file__)[0]),
        "demo_name": os.path.splitext(__file__)[0],
        "number_of_records": 1,
        "record_length": 300,
        "pretrig_samples": 8,
        "plot_fullscale": False,
        "input_range": [5000, 5000],
        "plot_show_pretrigger": True,
        "plot_disable_yticks": False,
        "plot_format": ".-",
        "plot_show_grid": True,
        "plot_trigger_at_zero": True,
        "input_impedance": [50] * len(enabled_channels),
        "trigger_mode": 23,
        "timeout": 4,
        "clock_source": 1,
        "enabled_channels": enabled_channels,
        "daisy_chain": deepcopy(daisy_chain_settings),
        "sample_skip_factor": 1,
    }

    try:
        (ADQAPI, adq_cu, nof_adqs, settings) = list_devices(settings)
    except NoDeviceFound as e:
        e.ADQAPI.DeleteADQControlUnit(e.adq_cu)
        raise

    adqs = [x for x in range(1, nof_adqs + 1)]

    # Detect daisy chain order
    (next_device, device_position, master_device) = detect_daisy_chain_order(
        ADQAPI, adq_cu, adqs
    )

    # input()

    print("Found master digitizer num: {}".format(master_device))
    print("Device chain")
    for adq, next_adq in zip(adqs, next_device):
        print(" {} -> {}".format(adq, next_adq))

    for idx, pos in enumerate(device_position):
        settings[idx]["daisy_chain"]["position"] = pos

    # settings[device_position.index(0)]["sdcard_enable"] = 1
    # settings[device_position.index(0)]["daisy_chain"]["compare_trig_info_sdcard"] = True
    # print("Blinking with master..")
    # ADQAPI.ADQ_Blink(adq_cu, master_device)

    print("Adjusting pretrigger..")
    try:
        for idx, adq_num in enumerate(adqs):
            additional_pretrigger_samples = get_daisy_chain_nof_pretrigger_samples(
                ADQAPI, adq_cu, adq_num, settings[idx]
            )
            settings[idx]["pretrig_samples"] += additional_pretrigger_samples
            print(
                "  Adding {} pretrigger samples to device in position {}.".format(
                    additional_pretrigger_samples,
                    settings[idx]["daisy_chain"]["position"],
                )
            )
    except Exception:
        close_all_devices(ADQAPI, adq_cu, nof_adqs)
        raise

    # Construct daisy chain device information objects and add to the settings dict
    for sett in settings:
        sett["daisy_chain"]["device_info"] = fill_daisy_chain_device_information(
            settings
        )

    print("Setting up collection..")
    try:
        for idx, adq_num in enumerate(adqs):
            setup_collection(ADQAPI, adq_cu, adq_num, settings[idx])
    except Exception:
        close_all_devices(ADQAPI, adq_cu, nof_adqs)
        raise

    print("Setting up daisy chain..")
    try:
        for idx, adq_num in enumerate(adqs):
            setup_daisy_chain(ADQAPI, adq_cu, adq_num, settings[idx])
    except Exception:
        close_all_devices(ADQAPI, adq_cu, nof_adqs)
        raise

    print("Setting up sdcard..")
    try:
        for idx, adq_num in enumerate(adqs):
            setup_sdcard(ADQAPI, adq_cu, adq_num, settings[idx])
    except Exception:
        close_all_devices(ADQAPI, adq_cu, nof_adqs)
        raise

    print("Arming devices..")
    try:
        arm_all(ADQAPI, adq_cu, adqs)
    except Exception:
        close_all_devices(ADQAPI, adq_cu, adqs)
        raise

    print("Enabling daisy chain..")
    try:
        enable_daisy_chain_all(ADQAPI, adq_cu, adqs, settings)
    except Exception:
        close_all_devices(ADQAPI, adq_cu, adqs)
        raise

    try:
        send_triggers(ADQAPI, adq_cu, adqs, settings)
    except Exception:
        close_all_devices(ADQAPI, adq_cu, adqs)
        raise

    print("Waiting for triggers..")
    try:
        wait_for_acquired_all(ADQAPI, adq_cu, adqs, settings)
    except Exception:
        close_all_devices(ADQAPI, adq_cu, adqs)
        raise

    print("Collecting data")
    try:
        data = collect_data_all(ADQAPI, adq_cu, adqs, settings)
    except Exception:
        close_all_devices(ADQAPI, adq_cu, adqs)
        raise

    save_data(data, settings)
    plot_daisy_chain(data, settings)


if __name__ == "__main__":
    main()
