#!/usr/bin/env python3
# vim:fileencoding=utf-8
"""
 Copyright 2018 Teledyne Signal Processing Devices Sweden AB
"""

import time
import numpy as np
import ctypes as ct
import matplotlib.pyplot as plt
import json
import os
import csv

from copy import deepcopy


class Error(Exception):
    pass


class NoDeviceFound(Error):
    def __init__(self, ADQAPI, adq_cu, *args, **kwargs):
        self.ADQAPI = ADQAPI
        self.adq_cu = adq_cu
        super().__init__(*args, **kwargs)


class DeviceFailedToStart(NoDeviceFound):
    pass


class HEADER(ct.Structure):
    _fields_ = [
        ("RecordStatus", ct.c_ubyte),
        ("UserID", ct.c_ubyte),
        ("Channel", ct.c_ubyte),
        ("DataFormat", ct.c_ubyte),
        ("SerialNumber", ct.c_uint32),
        ("RecordNumber", ct.c_uint32),
        ("SamplePeriod", ct.c_int32),
        ("Timestamp", ct.c_int64),
        ("RecordStart", ct.c_int64),
        ("RecordLength", ct.c_uint32),
        ("GPVector0", ct.c_uint16),
        ("GPVector1", ct.c_uint16),
    ]

    def to_dict(self):
        return {
            "RecordStatus": self.RecordStatus,
            "UserID": self.UserID,
            "Channel": self.Channel,
            "DataFormat": self.DataFormat,
            "SerialNumber": self.SerialNumber,
            "RecordNumber": self.RecordNumber,
            "SamplePeriod": self.SamplePeriod,
            "Timestamp": self.Timestamp,
            "RecordStart": self.RecordStart,
            "RecordLength": self.RecordLength,
            "GPVector0": self.GPVector0,
            "GPVector1": self.GPVector1,
        }


class TriggerInformation(ct.Structure):
    _fields_ = [
        ("Timestamp", ct.c_uint64),
        ("RecordStart", ct.POINTER(ct.c_int64)),
        ("ExtendedPrecision", ct.POINTER(ct.c_double)),
    ]

    def to_dict(self, length):
        record_start_list = []
        extended_precision_list = []
        for i in range(length):
            record_start_list.append(self.RecordStart[i])
            extended_precision_list.append(self.ExtendedPrecision[i])

        return {
            "Timestamp": self.Timestamp,
            "RecordStart": record_start_list,
            "ExtendedPrecision": extended_precision_list,
        }


class DeviceInformation(ct.Structure):
    _fields_ = [
        ("Position", ct.c_uint32),
        ("PretriggerSamples", ct.c_uint32),
        ("TriggerDelaySamples", ct.c_uint32),
    ]

    def to_dict(self):
        return {
            "position": self.Position,
            "pretrigger_samples": self.PretriggerSamples,
            "trigger_delay_samples": self.TriggerDelaySamples,
        }


def list_devices(settings=None):
    adqapi_paths = ["ADQAPI.dll"]
    adqapi_loaded = False
    for path in adqapi_paths:
        try:
            ADQAPI = ct.cdll.LoadLibrary(path)
            adqapi_loaded = True
            break
        except OSError:
            pass

    if not adqapi_loaded:
        raise Error("Failed to initialize the ADQAPI")

    # Manually set return type from some ADQAPI functions
    ADQAPI.CreateADQControlUnit.restype = ct.c_void_p
    ADQAPI.ADQ_GetRevision.restype = ct.c_void_p
    ADQAPI.ADQ_GetPtrStream.restype = ct.POINTER(ct.c_int16)
    ADQAPI.ADQControlUnit_FindDevices.argtypes = [ct.c_void_p]
    ADQAPI.ADQ_DebugCmd.argtypes = [
        ct.c_void_p,
        ct.c_void_p,
        ct.c_uint32,
        ct.c_uint32,
        ct.c_uint32,
        ct.c_uint32,
        ct.c_float,
        ct.c_void_p,
        ct.c_void_p,
        ct.c_void_p,
    ]
    ADQAPI.ADQ_GetBoardSerialNumber.restype = ct.c_char_p
    ADQAPI.ADQ_GetBoardProductName.restype = ct.c_char_p

    adq_cu = ct.c_void_p(ADQAPI.CreateADQControlUnit())

    ADQAPI.ADQControlUnit_EnableErrorTrace(adq_cu, 0x80000, ".")

    print("Enumerating devices..")
    ADQAPI.ADQControlUnit_FindDevices(adq_cu)

    nof_adqs = ADQAPI.ADQControlUnit_NofADQ(adq_cu)

    nof_failed_devices = ADQAPI.ADQControlUnit_GetFailedDeviceCount(adq_cu)

    if nof_failed_devices > 0:
        raise DeviceFailedToStart(
            ADQAPI, adq_cu, "Failed to start {} devices".format(nof_failed_devices)
        )

    if nof_adqs == 0:
        raise NoDeviceFound(ADQAPI, adq_cu, "No devices found")

    sn_list = []
    for num in range(1, nof_adqs + 1):
        dev_sn = ADQAPI.ADQ_GetBoardSerialNumber(adq_cu, num).decode("utf-8")
        print("Found device: {}".format(dev_sn))
        sn_list.append(dev_sn)

    if settings is not None:
        # Make one copy settings dictionary per device
        adqs = [x for x in range(1, nof_adqs + 1)]
        settings_list = [deepcopy(settings) for x in range(len(adqs))]

        # Add serial number to settings dict
        for sett, sn in zip(settings_list, sn_list):
            sett["dev_sn"] = sn
        for sett, adq in zip(settings_list, adqs):
            sett["adq_num"] = adq
        # Add PCIe address to settings dict
        for sett, adq in zip(settings_list, adqs):
            sett["pcie_addr"] = ADQAPI.ADQ_GetPCIeAddress(adq_cu, adq)

        return (ADQAPI, adq_cu, nof_adqs, settings_list)

    return (ADQAPI, adq_cu, nof_adqs)


def setup_buffers(number_of_records, record_length, number_of_channels):
    """
    Allocate data buffers
    """
    # Allocate target buffers for intermediate data storage
    target_buffers = (
        ct.POINTER(ct.c_int16 * (number_of_records * record_length))
        * number_of_channels
    )()

    for bufp in target_buffers:
        bufp.contents = (ct.c_int16 * (number_of_records * record_length))()

    # Allocate target buffers for headers
    header_list = (HEADER * number_of_records)()
    target_headers = ct.POINTER(HEADER * number_of_records)()
    target_headers.contents = header_list

    return (target_buffers, target_headers, header_list)


def setup_sdcard(ADQAPI, adq_cu, adq_num, settings):
    """
    Must be called after multi record configuration
    """

    sdcard_enable = settings.get("sdcard_enable", 0)

    if sdcard_enable:
        status = ADQAPI.ADQ_SDCardInit(adq_cu, adq_num)

        if not status:
            raise Error("SDCardInit failed")

        nof_sectors_to_erase = ADQAPI.ADQ_SDCardGetNofSectors(adq_cu, adq_num)

        status = ADQAPI.ADQ_SDCardErase(adq_cu, adq_num, 0, nof_sectors_to_erase)
        if not status:
            raise Error("SDCardErase failed")

        status = ADQAPI.ADQ_SDCardBackupSetAdditionalData(adq_cu, adq_num, 0)
        if not status:
            raise Error("SDCardBackupSetAdditionalData failed")

    status = ADQAPI.ADQ_SDCardBackupEnable(adq_cu, adq_num, sdcard_enable)
    if not status:
        raise Error("SDCardBackupEnable failed")


def setup_daisy_chain(ADQAPI, adq_cu, adq_num, settings):
    level_trig_channel = settings["daisy_chain"]["level_trig_channel"]
    trigger_source = settings["daisy_chain"]["trigger_source"]
    level_trig_edge = settings["daisy_chain"]["level_trig_edge"]
    level_trig_level = settings["daisy_chain"]["level_trig_level"]
    level_trig_arm_hyst = settings["daisy_chain"]["level_trig_arm_hyst"]
    position = settings["daisy_chain"]["position"]

    status = ADQAPI.ADQ_DaisyChainEnable(adq_cu, adq_num, 0)
    if not status:
        raise Error("EnableDaisyChain failed")

    status = ADQAPI.ADQ_DaisyChainReset(adq_cu, adq_num)
    if not status:
        raise Error("ResetDaisyChain failed")

    status = ADQAPI.ADQ_DaisyChainSetMode(adq_cu, adq_num, position == 0)
    if not status:
        raise Error("SetDaisyChainMode failed")

    if position == 0:
        status = ADQAPI.ADQ_DaisyChainSetupLevelTrigger(
            adq_cu,
            adq_num,
            level_trig_channel,
            level_trig_level,
            level_trig_arm_hyst,
            level_trig_edge,
        )
        if not status:
            raise Error("SetupDaisyChainLevelTrigger failed")

    status = ADQAPI.ADQ_DaisyChainSetupOutput(adq_cu, adq_num, 1, 0, 10 * 4)
    if not status:
        raise Error("SetupDaisyChainOutput failed")

    status = ADQAPI.ADQ_DaisyChainEnableOutput(adq_cu, adq_num, 1)
    if not status:
        raise Error("EnableDaisyChainOutput failed")

    status = ADQAPI.ADQ_DaisyChainSetTriggerSource(adq_cu, adq_num, trigger_source)
    if not status:
        raise Error("SetDaisyChainTriggerSource failed")


def detect_daisy_chain_order(ADQAPI, adq_cu, adqs):
    # First disable all outputs
    for adq_num in adqs:
        status = ADQAPI.ADQ_DaisyChainEnableOutput(adq_cu, adq_num, 0)
        if not status:
            raise Error("Disable daisy chain sync out failed")

        status = ADQAPI.ADQ_DaisyChainSetOutputState(adq_cu, adq_num, 0)
        if not status:
            raise Error("Setting daisy chain sync out state failed")

    next_device = [0] * len(adqs)
    sync_in_state = ct.c_uint32()
    for idx, adq_num in enumerate(adqs):
        status = ADQAPI.ADQ_DaisyChainEnableOutput(adq_cu, adq_num, 1)
        if not status:
            raise Error("Enable daisy chain sync out failed")

        status = ADQAPI.ADQ_DaisyChainSetOutputState(adq_cu, adq_num, 1)
        if not status:
            raise Error("Setting daisy chain sync out state failed")

        for adq_in_num in adqs:
            if adq_in_num == adq_num:
                continue
            status = ADQAPI.ADQ_DaisyChainGetInputState(
                adq_cu, adq_in_num, ct.byref(sync_in_state)
            )

            if sync_in_state.value:
                next_device[idx] = adq_in_num

        status = ADQAPI.ADQ_DaisyChainEnableOutput(adq_cu, adq_num, 0)
        if not status:
            raise Error("Disable daisy chain sync out failed")

        status = ADQAPI.ADQ_DaisyChainSetOutputState(adq_cu, adq_num, 0)
        if not status:
            raise Error("Setting daisy chain sync out state failed")

    nof_unconnected_slaves = sum(x == 0 for x in next_device)
    if nof_unconnected_slaves != 1:
        raise Error(
            "Too many unconnected sync out. Should be 1, is {}".format(
                nof_unconnected_slaves
            )
        )

    # Determine which digitizer is not pointed to by any device.
    # Filter out the last slave (next_device == 0)
    master_device = [x for x in (set(adqs) - set(next_device)) if x != 0]

    if len(master_device) != 1:
        raise Error(
            "Invalid master configuration. Found {} masters, should be 1".format(
                len(master_device)
            )
        )

    master_device = master_device[0]

    # Walk through the next_device list to determine the devices' positions in
    # the daisy chain.
    device_position = [0] * len(adqs)

    last_device_num = master_device
    for i in range(1, len(next_device)):
        next_slave_num = next_device[last_device_num - 1]
        device_position[next_slave_num - 1] = i

        last_device_num = next_slave_num

    return (next_device, device_position, master_device)


def get_daisy_chain_nof_pretrigger_samples(ADQAPI, adq_cu, adq_num, settings):
    position = settings["daisy_chain"]["position"]
    sample_skip_factor = settings.get("sample_skip_factor", 1)
    additional_pretrigger_samples = ct.c_uint32()

    status = ADQAPI.ADQ_DaisyChainGetNofPretriggerSamples(
        adq_cu,
        adq_num,
        position,
        sample_skip_factor,
        ct.byref(additional_pretrigger_samples),
    )
    if not status:
        raise Error("GetDaisyChainNofPretriggerSamples failed")

    return additional_pretrigger_samples.value


def get_dict_from_list(dict_list, val, key):
    return next((item for item in dict_list if item[key] == val), None)


def update_delays(settings, delay_list):

    for sett in settings:
        delay_dict = get_dict_from_list(delay_list, sett["dev_sn"], key="serial_number")
        sett["ref_clk_delay_adjust_ps"] = -delay_dict["value"]

    return settings


def setup_collection(ADQAPI, adq_cu, adq_num, settings):
    enabled_channels = settings["enabled_channels"]
    trigger_mode = settings["trigger_mode"]
    pretrig_samples = settings["pretrig_samples"]
    number_of_records = settings["number_of_records"]
    record_length = settings["record_length"]
    input_impedance = settings.get("input_impedance", [None] * len(enabled_channels))
    input_range = settings.get("input_range", [None] * len(enabled_channels))
    dc_offset = settings.get("dc_offset", [None] * len(enabled_channels))
    clock_source = settings.get("clock_source", None)
    ref_clock_delay_ps = settings.get("ref_clk_delay_adjust_ps", None)
    sample_skip_factor = settings.get("sample_skip_factor", None)
    trigger_delay_samples = settings.get("trigger_delay_samples", None)

    level_trig_channel = settings.get("level_trig_channel", None)
    level_trig_edge = settings.get("level_trig_edge", None)
    level_trig_level = settings.get("level_trig_level", None)
    level_trig_arm_hyst = settings.get("level_trig_arm_hyst", None)

    do_sleep = False

    channel_mask = 0
    for ch in enabled_channels:
        channel_mask |= 1 << (ch - 1)

    for chidx, ch in enumerate(enabled_channels):
        if input_range[chidx] is not None:
            status = ADQAPI.ADQ_SetInputRange(
                adq_cu, adq_num, ch, ct.c_float(input_range[chidx]), None
            )
            if not status:
                raise Error("SetInputRange failed")
    for chidx, ch in enumerate(enabled_channels):
        if dc_offset[chidx] is not None:
            status = ADQAPI.ADQ_SetAdjustableBias(
                adq_cu, adq_num, ch, int(dc_offset[chidx] * 32767)
            )
            if not status:
                raise Error("SetAdjustableBias failed")
            do_sleep = True
    for chidx, ch in enumerate(enabled_channels):
        if input_impedance[chidx] == "50" or input_impedance[chidx] == 50:
            status = ADQAPI.ADQ_SetInputImpedance(adq_cu, adq_num, ch, 0)
            if not status:
                raise Error("SetInputImpedance failed")
        elif input_impedance[chidx] == "1M":
            status = ADQAPI.ADQ_SetInputImpedance(adq_cu, adq_num, ch, 1)
            if not status:
                raise Error("SetInputImpedance failed")
        elif input_impedance[chidx] is not None:
            raise Error(
                'Unsupported input impedance {}. Valid values are "50", "1M"'.format(
                    chidx
                )
            )

    if do_sleep:
        time.sleep(0.1)

    if ref_clock_delay_ps is not None:
        dev_sn = ADQAPI.ADQ_GetBoardSerialNumber(adq_cu, adq_num).decode("utf-8")
        print(
            "Adjusting reference clock delay for device {} to {} ps.".format(
                dev_sn, ref_clock_delay_ps
            )
        )
        status = ADQAPI.ADQ_AdjustClockReferenceDelay(
            adq_cu, adq_num, ref_clock_delay_ps
        )
        if not status:
            raise Error("AdjustClockReferenceDelay failed")

    if clock_source is not None:
        status = ADQAPI.ADQ_SetClockSource(adq_cu, adq_num, clock_source)
        if not status:
            raise Error("SetClockSource failed")

    if level_trig_level is not None:
        channel_mask = 1 << (level_trig_channel - 1)
        level = (ct.c_int * 8)(level_trig_level)
        edge = (ct.c_int * 8)(level_trig_edge)
        arm_hyst = (ct.c_int * 8)(level_trig_arm_hyst)
        status = ADQAPI.ADQ_SetupLevelTrigger(
            adq_cu,
            adq_num,
            ct.byref(level),
            ct.byref(edge),
            ct.byref(arm_hyst),
            channel_mask,
            1,
        )
        if not status:
            raise Error("SetupLevelTrigger failed")

    status = ADQAPI.ADQ_SetTriggerMode(adq_cu, adq_num, trigger_mode)
    if not status:
        raise Error("SetTriggerMode failed")

    status = ADQAPI.ADQ_SetPreTrigSamples(adq_cu, adq_num, pretrig_samples)
    if not status:
        raise Error("SetPreTrigSamples failed")

    if sample_skip_factor is not None:
        status = ADQAPI.ADQ_SetSampleSkip(adq_cu, adq_num, sample_skip_factor)
        if not status:
            raise Error("SetSampleSkip failed")

    if trigger_delay_samples is not None:
        status = ADQAPI.ADQ_SetTriggerHoldOffSamples(
            adq_cu, adq_num, trigger_delay_samples
        )
        if not status:
            raise Error("SetTriggerHoldOffSamples failed")

    status = ADQAPI.ADQ_MultiRecordSetChannelMask(adq_cu, adq_num, channel_mask)
    if not status:
        raise Error("MultiRecordSetChannelMask failed")

    status = ADQAPI.ADQ_MultiRecordSetup(
        adq_cu, adq_num, number_of_records, record_length
    )
    if not status:
        raise Error("MultiRecordSetup failed")


def wait_for_acquired_all(ADQAPI, adq_cu, nof_adqs, settings):
    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    # Use timeout setting for channel 1
    if isinstance(settings, list):
        timeout = settings[0]["timeout"]
    else:
        timeout = settings["timeout"]

    start = time.time()
    acquired = [0] * len(adqs)
    while time.time() - start < timeout and not all(acquired):
        for adq_num in adqs:
            acquired[adq_num - 1] = ADQAPI.ADQ_GetAcquiredAll(adq_cu, adq_num)

    for num, acq in enumerate(acquired):
        if not acq:
            raise Error("Acquisition timedout for device {}".format(num + 1))

    for adq_num in adqs:
        if isinstance(settings, list):
            sett = settings[adq_num - 1]
            sdcard_enable = sett.get("sdcard_enable", 0)
        else:
            sdcard_enable = settings.get("sdcard_enable", 0)

        if not sdcard_enable:
            continue

        backup_status = ct.c_uint32(0)
        if not ADQAPI.ADQ_SDCardBackupGetStatus(
            adq_cu, adq_num, ct.byref(backup_status)
        ):
            raise Error("SDCardBackupGetStatus failed")

        while backup_status.value & (1 << 16):
            if not ADQAPI.ADQ_SDCardBackupGetStatus(
                adq_cu, adq_num, ct.byref(backup_status)
            ):
                raise Error("SDCardBackupGetStatus failed")
            time.sleep(0.1)

            # if not ADQAPI.ADQ_SDCardBackupGetProgress(adq_cu, adq_num, ct.byref(backup_progress)):
            #    raise Error("SDCardBackupGetProgress failed")

            # print("SD card backup progress {}".format(backup_progress.value))

        if not ADQAPI.ADQ_SDCardBackupGetStatus(
            adq_cu, adq_num, ct.byref(backup_status)
        ):
            raise Error("SDCardBackupGetStatus failed")

        if backup_status.value & (1 << 18):
            raise Error("SD card write failed")


def fill_daisy_chain_device_information(settings):

    # Put together device information.
    device_info = []
    for sett in settings:
        tmp = {}
        tmp["position"] = sett["daisy_chain"]["position"]
        tmp["pretrigger_samples"] = sett["pretrig_samples"]
        tmp["trigger_delay_samples"] = 0
        device_info.append(tmp)

    return device_info


def collect_data(ADQAPI, adq_cu, adq_num, settings):
    enabled_channels = settings["enabled_channels"]
    number_of_records = settings["number_of_records"]
    record_length = settings["record_length"]
    sample_skip_factor = settings.get("sample_skip_factor", 1)
    daisy_chain_settings = settings.get("daisy_chain", None)

    if daisy_chain_settings is not None:
        compare_trig_info_sdcard = daisy_chain_settings.get(
            "compare_trig_info_sdcard", False
        )
    else:
        compare_trig_info_sdcard = False

    number_of_channels = 8

    channel_mask = 0
    for ch in enabled_channels:
        channel_mask |= 1 << (ch - 1)

    (target_buffers, target_headers, header_list) = setup_buffers(
        number_of_records, record_length, number_of_channels
    )

    # Retrieve daisy chain trigger information.
    daisy_chain_trig_info = []
    if daisy_chain_settings is not None:
        if daisy_chain_settings["position"] == 0:
            # Only fetch information from the master device.
            device_info = daisy_chain_settings["device_info"]
            nof_devices = len(device_info)

            device_info_array = (DeviceInformation * nof_devices)()
            for idx, info in enumerate(device_info):
                device_info_array[idx].Position = info["position"]
                device_info_array[idx].PretriggerSamples = info["pretrigger_samples"]
                device_info_array[idx].TriggerDelaySamples = info[
                    "trigger_delay_samples"
                ]

            trig_info = (TriggerInformation * number_of_records)()
            record_start_array_list = []
            extended_precision_array_list = []
            for i in range(number_of_records):
                record_start_array_list.append((ct.c_int64 * nof_devices)())
                extended_precision_array_list.append((ct.c_double * nof_devices)())
                trig_info[i].Timestamp = 0
                trig_info[i].RecordStart = ct.cast(
                    ct.pointer(record_start_array_list[-1]), ct.POINTER(ct.c_int64)
                )
                trig_info[i].ExtendedPrecision = ct.cast(
                    ct.pointer(extended_precision_array_list[-1]),
                    ct.POINTER(ct.c_double),
                )

            status = ADQAPI.ADQ_DaisyChainGetTriggerInformation(
                adq_cu,
                adq_num,
                daisy_chain_settings["trigger_source"],
                daisy_chain_settings["level_trig_edge"],
                daisy_chain_settings["level_trig_level"],
                daisy_chain_settings["level_trig_channel"],
                number_of_records,
                record_length,
                sample_skip_factor,
                ct.byref(device_info_array),
                len(device_info),
                ct.byref(trig_info),
            )
            if not status:
                raise Error(
                    "Failed to get daisy chain trigger information for master device {}".format(
                        adq_num
                    )
                )

            if compare_trig_info_sdcard:
                print("Reading trig info from SD card")
                trig_info_sdcard = (TriggerInformation * number_of_records)()
                record_start_array_list_sdcard = []
                extended_precision_array_list_sdcard = []
                for i in range(number_of_records):
                    record_start_array_list_sdcard.append((ct.c_int64 * nof_devices)())
                    extended_precision_array_list_sdcard.append(
                        (ct.c_double * nof_devices)()
                    )
                    trig_info_sdcard[i].Timestamp = 0
                    trig_info_sdcard[i].RecordStart = ct.cast(
                        ct.pointer(record_start_array_list_sdcard[-1]),
                        ct.POINTER(ct.c_int64),
                    )
                    trig_info_sdcard[i].ExtendedPrecision = ct.cast(
                        ct.pointer(extended_precision_array_list_sdcard[-1]),
                        ct.POINTER(ct.c_double),
                    )
                status = ADQAPI.ADQ_SDCardBackupDaisyChainGetTriggerInformation(
                    adq_cu,
                    adq_num,
                    daisy_chain_settings["trigger_source"],
                    daisy_chain_settings["level_trig_edge"],
                    daisy_chain_settings["level_trig_level"],
                    daisy_chain_settings["level_trig_channel"],
                    number_of_records,
                    record_length,
                    ct.byref(device_info_array),
                    len(device_info),
                    ct.byref(trig_info_sdcard),
                )
                if not status:
                    raise Error(
                        "Failed to get daisy chain trigger "
                        "information for master device {}".format(adq_num)
                    )

                for trig_info_record, trig_info_record_sdcard in zip(
                    trig_info, trig_info_sdcard
                ):
                    trig_info_record = trig_info_record.to_dict(nof_devices)
                    trig_info_record_sdcard = trig_info_record_sdcard.to_dict(
                        nof_devices
                    )

                    if (
                        trig_info_record["Timestamp"]
                        != trig_info_record_sdcard["Timestamp"]
                    ):
                        raise Error(
                            "Timestamp mismatch between SD card "
                            "and DRAM for DC trigger info"
                        )

                    if (
                        trig_info_record["RecordStart"]
                        != trig_info_record_sdcard["RecordStart"]
                    ):
                        raise Error(
                            "Record start mismatch between SD card "
                            "and DRAM for DC trigger info"
                        )

                    if (
                        trig_info_record["ExtendedPrecision"]
                        != trig_info_record_sdcard["ExtendedPrecision"]
                    ):
                        raise Error(
                            "Extended precision mismatch between SD card "
                            "and DRAM for DC trigger info"
                        )

            trig_info_list = [x.to_dict(nof_devices) for x in trig_info]
            for info in trig_info_list:
                daisy_chain_trig_info.append(
                    {
                        "timestamp": info["Timestamp"],
                        "record_start": info["RecordStart"],
                        "extended_precision": info["ExtendedPrecision"],
                    }
                )

    records_completed = 0
    status = ADQAPI.ADQ_GetDataWHTS(
        adq_cu,
        adq_num,
        target_buffers,
        target_headers,
        None,
        number_of_records * record_length,
        2,
        records_completed,
        number_of_records,
        channel_mask,
        0,
        record_length,
        0x00,
    )

    if not status:
        raise Error("Failed to get data for device {}".format(adq_num))

    data_16bit = [
        np.frombuffer(
            target_buffers[ch - 1].contents,
            dtype=np.int16,
            count=record_length * number_of_records,
        )
        for ch in enabled_channels
    ]

    header_list = [x.to_dict() for x in header_list]

    return (header_list, data_16bit, daisy_chain_trig_info)


def collect_data_all(ADQAPI, adq_cu, nof_adqs, settings):

    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    data = [None] * len(adqs)
    for adq_num in adqs:
        d = {"adq_num": adq_num}

        if isinstance(settings, list):
            (d["headers"], d["data"], d["daisy_chain_trig_info"]) = collect_data(
                ADQAPI, adq_cu, adq_num, settings[adq_num - 1]
            )
        else:
            (d["headers"], d["data"], d["daisy_chain_trig_info"]) = collect_data(
                ADQAPI, adq_cu, adq_num, settings
            )

        data[adq_num - 1] = d

    return data


def send_triggers(ADQAPI, adq_cu, nof_adqs, settings):
    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    if isinstance(settings, dict):
        settings = [settings] * len(adqs)

    for adq_num, sett in zip(adqs, settings):
        daisy_chain = sett.get("daisy_chain", None)
        if sett["trigger_mode"] == 23:
            if daisy_chain is not None and daisy_chain["trigger_source"] != 1:
                return
        elif sett["trigger_mode"] != 1:
            return

        for rec in range(sett["number_of_records"]):
            status = ADQAPI.ADQ_SWTrig(adq_cu, adq_num)
            if not status:
                raise Error(
                    "Failed to send software trigger for device {}".format(adq_num)
                )
            time.sleep(0.05)


def arm_all(ADQAPI, adq_cu, nof_adqs):
    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    for adq_num in adqs:
        status = ADQAPI.ADQ_DisarmTrigger(adq_cu, adq_num)
        if not status:
            raise Error("Failed to disarm trigger for device  {}".format(adq_num))
    for adq_num in adqs:
        status = ADQAPI.ADQ_ArmTrigger(adq_cu, adq_num)
        if not status:
            raise Error("Failed to disarm trigger for device  {}".format(adq_num))


def enable_daisy_chain_all(ADQAPI, adq_cu, nof_adqs, settings):
    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    # Enable slaves
    for adq_num, sett in zip(adqs, settings):
        if sett["daisy_chain"]["position"] == 0:
            continue
        status = ADQAPI.ADQ_DaisyChainEnable(adq_cu, adq_num, 1)
        if not status:
            raise Error("Failed to enable daisy chain for device {}".format(adq_num))

    found_master = False
    # Enable master
    for adq_num, sett in zip(adqs, settings):
        if sett["daisy_chain"]["position"] == 0:
            if found_master:
                raise Error("Two daisy chain masters found.")

            status = ADQAPI.ADQ_DaisyChainEnable(adq_cu, adq_num, 1)
            if not status:
                raise Error(
                    "Failed to enable daisy chain for device {}".format(adq_num)
                )
            found_master = True

    if not found_master:
        raise Error("No daisy chain master device found.")


def setup_filter(ADQAPI, adq_cu, adq_num, settings):
    enabled_channels = settings["enabled_channels"]
    coeffs = settings["filter_settings"]["coeffs"]

    for chidx, ch in enumerate(enabled_channels):
        status = ADQAPI.ADQ_ResetUserLogicFilter(adq_cu, adq_num, ch)
        if not status:
            raise Error("ResetUserLogicFilter failed")

        ch_coeffs = coeffs[chidx]
        ct_coeffs = (ct.c_float * len(ch_coeffs))(*ch_coeffs)
        status = ADQAPI.ADQ_SetUserLogicFilter(
            adq_cu, adq_num, ch, ct.byref(ct_coeffs), len(ch_coeffs), 1, 0
        )
        if not status:
            raise Error("SetUserLogicFilter failed")

        status = ADQAPI.ADQ_EnableUserLogicFilter(adq_cu, adq_num, ch, 1)
        if not status:
            raise Error("EnableUserLogicFilter failed")


def close_all_devices(ADQAPI, adq_cu, nof_adqs):
    if isinstance(nof_adqs, list):
        adqs = nof_adqs
    else:
        adqs = [x for x in range(1, nof_adqs + 1)]

    for adq_num in adqs:
        ADQAPI.ADQ_DisarmTrigger(adq_cu, adq_num)
        ADQAPI.ADQ_MultiRecordClose(adq_cu, adq_num)


def plot_combined(ddict, settings):
    if isinstance(settings, dict):
        settings = [settings] * len(ddict)

    enabled_channels = settings[0]["enabled_channels"]
    plot_fullscale = settings[0].get("plot_fullscale", False)
    plot_show_pretrigger = settings[0].get("plot_show_pretrigger", False)
    plot_disable_yticks = settings[0].get("plot_disable_yticks", True)
    plot_format = settings[0].get("plot_format", "")
    plot_show_grid = settings[0].get("plot_show_grid", False)

    plot_adq_map = settings[0].get("plot_adq_map", None)  # [2, 1, 3, 0]

    if plot_adq_map is not None:
        ddict = [ddict[i] for i in plot_adq_map]

    nof_adqs = len(ddict)

    fig, ax = plt.subplots(len(enabled_channels), nof_adqs, sharey=True, sharex=True)
    if nof_adqs == 1:
        ax = [[a] for a in ax]

    for adq_index in range(len(ddict)):
        sett = settings[adq_index]
        data = ddict[adq_index]
        headers = data["headers"]
        data_all_channels = data["data"]

        ax[0][adq_index].set_title("Device SPD-{}".format(headers[0]["SerialNumber"]))

        for ch, data_channel in enumerate(data_all_channels):
            x_start = 0
            for rec_num, h in enumerate(headers):
                # Only plot channel 0 for now
                record_length = h["RecordLength"]
                pretrig_samples = sett["pretrig_samples"]

                if rec_num > 0:
                    ax[ch][adq_index].axvline(x_start, 0, 1, color="C2")

                if pretrig_samples > 0 and plot_show_pretrigger:
                    y = data_channel[x_start : x_start + pretrig_samples + 1]
                    if plot_fullscale:
                        y = y / 32768
                    x = x_start + np.arange(pretrig_samples + 1)
                    ax[ch][adq_index].plot(
                        x, y, plot_format, color="C1", label="Pre-trigger"
                    )
                    x_start += pretrig_samples

                    record_length -= pretrig_samples

                    color = "C0"
                else:
                    color = "C{}".format(rec_num)

                y = data_channel[x_start : x_start + record_length]
                if plot_fullscale:
                    y = y / 32768
                x = x_start + np.arange(record_length)
                ax[ch][adq_index].plot(x, y, plot_format, color=color)

                if plot_fullscale:
                    ax[ch][adq_index].set_ylim((-1, 1))
                else:
                    ax[ch][adq_index].set_ylim((-32768, 32768))

            if plot_show_grid:
                ax[ch][adq_index].grid(which="both")
            if plot_disable_yticks:
                ax[ch][adq_index].set_yticks([])
                ax[ch][adq_index].set_xticks([])
    plt.show()


def plot_daisy_chain(ddict, settings):
    if isinstance(settings, dict):
        settings = [settings] * len(ddict)

    enabled_channels = settings[0]["enabled_channels"]
    plot_fullscale = settings[0].get("plot_fullscale", False)
    plot_disable_yticks = settings[0].get("plot_disable_yticks", True)
    plot_format = settings[0].get("plot_format", "")
    plot_show_grid = settings[0].get("plot_show_grid", False)

    plot_adq_map = settings[0].get("plot_adq_map", None)

    if plot_adq_map is not None:
        ddict = [ddict[i] for i in plot_adq_map]
        settings = [settings[i] for i in plot_adq_map]

    nof_adqs = len(ddict)

    fig, ax = plt.subplots(len(enabled_channels), nof_adqs, sharey=True, sharex=True)
    if nof_adqs == 1:
        ax = [[a] for a in ax]

    # TODO: Have a single, common trig info dict.
    for sett in settings:
        if sett["daisy_chain"]["position"] == 0:
            master_index = settings.index(sett)
            break

    for adq_index in range(len(ddict)):
        sett = settings[adq_index]
        data = ddict[adq_index]
        headers = data["headers"]
        data_all_channels = data["data"]

        ax[0][adq_index].set_title("Device SPD-{}".format(headers[0]["SerialNumber"]))

        for ch, data_channel in enumerate(data_all_channels):
            x_start = 0
            for rec_num, h in enumerate(headers):
                # Only plot channel 0 for now
                record_length = h["RecordLength"]

                offset = (
                    ddict[master_index]["daisy_chain_trig_info"][rec_num][
                        "record_start"
                    ][int(data["adq_num"]) - 1]
                    + ddict[master_index]["daisy_chain_trig_info"][rec_num][
                        "extended_precision"
                    ][int(data["adq_num"]) - 1]
                ) / h["SamplePeriod"]

                y = data_channel[x_start : x_start + record_length]
                if plot_fullscale:
                    y = y / 32768

                x = offset + np.arange(len(y))

                x_start += record_length
                ax[ch][adq_index].plot(x, y, plot_format)

                if plot_fullscale:
                    ax[ch][adq_index].set_ylim((-1, 1))
                else:
                    ax[ch][adq_index].set_ylim((-32768, 32768))

            if plot_show_grid:
                ax[ch][adq_index].grid(which="both")
            if plot_disable_yticks:
                ax[ch][adq_index].set_yticks([])
                ax[ch][adq_index].set_xticks([])

    plt.show()


def plot_data(ddict, settings):
    if isinstance(settings, dict):
        settings = [settings] * len(ddict)

    for sett, data in zip(settings, ddict):
        enabled_channels = sett["enabled_channels"]
        plot_vertical = sett.get("plot_vertical", True)
        plot_fullscale = sett.get("plot_fullscale", False)
        plot_show_pretrigger = sett.get("plot_show_pretrigger", False)
        plot_channels = sett.get("plot_channels", enabled_channels)
        plot_format = sett.get("plot_format", "")
        plot_show_grid = sett.get("plot_show_grid", False)
        input_impedance = sett.get("input_impedance", [None] * len(enabled_channels))
        input_range = sett.get("input_range", [None] * len(enabled_channels))
        dc_offset = sett.get("dc_offset", [None] * len(enabled_channels))

        headers = data["headers"]
        data_all_channels = data["data"]
        if plot_vertical:
            fig, ax = plt.subplots(len(plot_channels), 1, sharey=True, sharex=True)
        else:
            fig, ax = plt.subplots(1, len(plot_channels), sharey=True, sharex=True)

        if len(plot_channels) == 1:
            ax = [ax]

        sup_title = ""
        if input_range[1:] == input_range[:-1] and input_range[0] is not None:
            sup_title += "{:.2} Vpp, ".format(input_range[0] / 1000)

        if dc_offset[1:] == dc_offset[:-1] and dc_offset[0] is not None:
            sup_title += "{:.2} V DC offset, ".format(float(dc_offset[0]))

        if (
            input_impedance[1:] == input_impedance[:-1]
            and input_impedance[0] is not None
        ):
            sup_title += "{} Ohm".format(input_impedance[0])

        sup_title = "{}: {}".format(sett["demo_name"], sup_title)

        fig.suptitle(sup_title)
        fig.tight_layout()

        data_selected_channels = [
            data_all_channels[enabled_channels.index(ch)] for ch in plot_channels
        ]

        for ch, data_channel in enumerate(data_selected_channels):
            x_start = 0
            for rec_num, h in enumerate(headers):
                # Only plot channel 0 for now
                record_length = h["RecordLength"]
                pretrig_samples = sett["pretrig_samples"]

                if rec_num > 0:
                    ax[ch].axvline(x_start, 0, 1, color="C2")

                if pretrig_samples > 0 and plot_show_pretrigger:
                    y = data_channel[x_start : x_start + pretrig_samples]
                    if plot_fullscale:
                        y = y / 32768
                    x = x_start + np.arange(pretrig_samples)
                    ax[ch].plot(x, y, plot_format, color="C1", label="Pre-trigger")
                    x_start += pretrig_samples

                    record_length -= pretrig_samples

                    color = "C0"
                else:
                    color = "C{}".format(rec_num)

                y = data_channel[x_start : x_start + record_length]
                if plot_fullscale:
                    y = y / 32768
                x = x_start + np.arange(record_length)
                ax[ch].plot(x, y, plot_format, color=color)

                title = ""
                if input_range[1:] != input_range[:-1]:
                    title += "{:.2} Vpp ".format(input_range[ch] / 1000)
                if dc_offset[1:] != dc_offset[:-1]:
                    title += "{:.2} V DC offset ".format(dc_offset[ch])
                if input_impedance[1:] != input_impedance[:-1]:
                    title += "{} Ohm".format(input_impedance[ch])

                if len(title) > 0:
                    ax[ch].set_title(title.format(dc_offset[ch], input_impedance[ch]))
                x_start += record_length

                box = ax[ch].get_position()
                ax[ch].set_position([box.x0, box.y0, box.width, box.height * 0.9])

                if ch == 0:
                    if pretrig_samples > 0 and plot_show_pretrigger:
                        ax[ch].legend(
                            ["Pre-trigger", "Post-trigger"],
                            loc="lower left",
                            bbox_to_anchor=(0, 1),
                            fancybox=True,
                        )

                if ch == 0 or plot_vertical:
                    if plot_fullscale:
                        ax[ch].set_ylabel("Full-scale")
                    else:
                        ax[ch].set_ylabel("ADC Codes")

                if ch == (len(data_all_channels) - 1) or not plot_vertical:
                    ax[ch].set_xlabel("Time [ns]")

                if plot_fullscale:
                    ax[ch].set_ylim((-1, 1))
                else:
                    ax[ch].set_ylim((-32768, 32768))

                ax[ch].set_xlim(0, record_length - 1)
                if plot_show_grid:
                    ax[ch].grid(which="both")

    plt.show()


def plot_superpos(ddict, settings):
    if isinstance(settings, dict):
        settings = [settings] * len(ddict)

    for sett, data in zip(settings, ddict):
        enabled_channels = sett["enabled_channels"]
        plot_vertical = sett.get("plot_vertical", True)
        plot_fullscale = sett.get("plot_fullscale", False)
        plot_channels = sett.get("plot_channels", enabled_channels)
        plot_format = sett.get("plot_format", "")
        plot_show_grid = sett.get("plot_show_grid", False)
        input_impedance = sett.get("input_impedance", [None] * len(enabled_channels))
        input_range = sett.get("input_range", [None] * len(enabled_channels))
        dc_offset = sett.get("dc_offset", [None] * len(enabled_channels))

        headers = data["headers"]
        data_all_channels = data["data"]

        fig, ax = plt.subplots(1, 1, sharey=True, sharex=True)

        sup_title = ""
        if input_range[1:] == input_range[:-1] and input_range[0] is not None:
            sup_title += "{:.2} Vpp, ".format(input_range[0] / 1000)

        if dc_offset[1:] == dc_offset[:-1] and dc_offset[0] is not None:
            sup_title += "{:.2} V DC offset, ".format(float(dc_offset[0]))

        if (
            input_impedance[1:] == input_impedance[:-1]
            and input_impedance[0] is not None
        ):
            sup_title += "{} Ohm".format(input_impedance[0])

        sup_title = "{}: {}".format(sett["demo_name"], sup_title)

        fig.suptitle(sup_title)
        fig.tight_layout()

        data_selected_channels = [
            data_all_channels[enabled_channels.index(ch)] for ch in plot_channels
        ]

        for ch, data_channel in enumerate(data_selected_channels):
            x_start = 0
            for rec_num, h in enumerate(headers):
                # Only plot channel 0 for now
                record_length = h["RecordLength"]

                y = data_channel[x_start : x_start + record_length]
                if plot_fullscale:
                    y = y / 32768
                x = x_start + np.arange(record_length)
                ax.plot(x, y, plot_format)

                title = ""
                if input_range[1:] != input_range[:-1]:
                    title += "{:.2} Vpp ".format(input_range[ch] / 1000)
                if dc_offset[1:] != dc_offset[:-1]:
                    title += "{:.2} V DC offset ".format(dc_offset[ch])
                if input_impedance[1:] != input_impedance[:-1]:
                    title += "{} Ohm".format(input_impedance[ch])

                if len(title) > 0:
                    ax.set_title(title.format(dc_offset[ch], input_impedance[ch]))
                x_start += record_length

                # box = ax.get_position()
                # ax.set_position([box.x0, box.y0, box.width, box.height * 0.9])

                if ch == 0:
                    if plot_fullscale:
                        ax.set_ylabel("Full-scale")
                    else:
                        ax.set_ylabel("ADC Codes")

                if ch == (len(data_all_channels) - 1) or not plot_vertical:
                    ax.set_xlabel("Time [ns]")

                if plot_fullscale:
                    ax.set_ylim((-1, 1))
                else:
                    ax.set_ylim((-32768, 32768))

                ax.set_xlim(0, record_length - 1)
                if plot_show_grid:
                    ax.grid(which="both")
            ax.legend(
                ["Ch {}".format(x) for x in enabled_channels],
                # loc="lower left",
                # bbox_to_anchor=(0, 1),
                fancybox=True,
            )

    plt.show()


def save_data(ddict, settings):

    if isinstance(settings, dict):
        settings = [settings] * len(ddict)

    for sett, data in zip(settings, ddict):
        data_path = sett.get("data_path", None)
        if data_path is not None:
            os.makedirs(data_path, exist_ok=True)
        else:
            data_path = "./"

        filename_data = "data_adq{}.bin".format(data["adq_num"])
        filename_header = "header_adq{}.json".format(data["adq_num"])
        filename_settings = "settings_adq{}.json".format(data["adq_num"])
        filename_trig_info = "trig_info_adq{}.json".format(data["adq_num"])

        filepath_data = os.path.join(data_path, filename_data)
        filepath_header = os.path.join(data_path, filename_header)
        filepath_settings = os.path.join(data_path, filename_settings)
        filepath_trig_info = os.path.join(data_path, filename_trig_info)

        data_all_channels = data["data"]
        data_len = len(data_all_channels[0])
        data_tot = np.zeros(sum([len(x) for x in data_all_channels]))
        for ch, d in enumerate(data_all_channels):
            data_tot[ch * data_len : (ch + 1) * data_len] = d

        data_tot.astype("int16").tofile(filepath_data)

        with open(filepath_header, "w") as fp:
            json.dump(data["headers"], fp, indent=4)

        with open(filepath_settings, "w") as fp:
            json.dump(sett, fp, indent=4)

        with open(filepath_trig_info, "w") as fp:
            json.dump(data["daisy_chain_trig_info"], fp, indent=4)


def read_data(path, adq_num):
    filename_data = "data_adq{}.bin".format(adq_num)
    filename_header = "header_adq{}.json".format(adq_num)
    filename_settings = "settings_adq{}.json".format(adq_num)
    filename_trig_info = "trig_info_adq{}.json".format(adq_num)

    filepath_data = os.path.join(path, filename_data)
    filepath_header = os.path.join(path, filename_header)
    filepath_settings = os.path.join(path, filename_settings)
    filepath_trig_info = os.path.join(path, filename_trig_info)

    with open(filepath_header, "r") as fp:
        header_list = json.load(fp)

    with open(filepath_settings, "r") as fp:
        settings = json.load(fp)

    with open(filepath_trig_info, "r") as fp:
        trig_info = json.load(fp)

    data_tot = np.fromfile(filepath_data, dtype="int16")

    nof_channels = len(settings["enabled_channels"])
    data_all_channels = [None] * nof_channels

    # Assume the same amount of data on each channel
    data_len = len(data_tot) // nof_channels
    for ch in range(nof_channels):
        data_all_channels[ch] = data_tot[ch * data_len : (1 + ch) * data_len]
        data = {
            "headers": header_list,
            "data": data_all_channels,
            "adq_num": adq_num,
            "daisy_chain_trig_info": trig_info,
        }

    return (settings, data)


def read_data_all(path, adq_nums):
    data = []
    settings = []

    # Settings is the same for all channels
    for num in sorted(adq_nums):
        stmp, dtmp = read_data(path, num)
        data.append(dtmp)
        settings.append(stmp)

    return (settings, data)


def read_delays(file_path):
    delays = []
    with open(file_path, "r", newline="") as fp:
        csv_reader = csv.reader(fp, delimiter=",")
        for line_count, row in enumerate(csv_reader):
            if line_count == 0:
                header = row
            else:
                # Ignore empty lines
                if len(row) == 0:
                    continue
                entry = {}
                for k, v in zip(header, row):
                    key = k.lower().replace(" ", "_")
                    if key == "value":
                        entry[key] = float(v)
                    else:
                        entry[key] = v

                delays += [entry]
    return delays
