/*
 *   Copyright 2018 Teledyne Signal Processing Devices Sweden AB
 */

#include <cstdlib>
#include <ctime>
#include <iostream>
#include <mutex>
#include <sstream>
#include <string>

#include <ADQAPI.h>
#include <nds3/nds.h>

#include "ADQAIChannel.h"
#include "ADQAIChannelGroup.h"
#include "ADQDefinition.h"
#include "ADQDevice.h"
#include "ADQInit.h"

ADQDevice::ADQDevice(const std::string& name, nds::Node& parentNode, ADQInterface* adqInterface) :
    m_productNamePV(nds::PVDelegateIn<std::string>("ProdName", std::bind(&ADQDevice::getProductName, this,
                                                                         std::placeholders::_1, std::placeholders::_2))),
    m_serialNumberPV(nds::PVDelegateIn<std::string>("ProdSerial", std::bind(&ADQDevice::getSerialNumber, this,
                                                                            std::placeholders::_1, std::placeholders::_2))),
    m_productIDPV(nds::PVDelegateIn<int32_t>("ProdID", std::bind(&ADQDevice::getProductID, this, std::placeholders::_1,
                                                                 std::placeholders::_2))),
    m_adqTypePV(nds::PVDelegateIn<int32_t>("ProdType", std::bind(&ADQDevice::getADQType, this, std::placeholders::_1,
                                                                 std::placeholders::_2))),
    m_cardOptionPV(nds::PVDelegateIn<std::string>("ProdOpt", std::bind(&ADQDevice::getCardOption, this,
                                                                       std::placeholders::_1, std::placeholders::_2))),
    m_isAlivePV(nds::PVDelegateIn<int32_t>("isAlive", std::bind(&ADQDevice::isAlive, this,
        std::placeholders::_1, std::placeholders::_2))),
    m_tempLocalPV(nds::PVDelegateIn<int32_t>("TempLocal", std::bind(&ADQDevice::getTempLocal, this,
        std::placeholders::_1, std::placeholders::_2))),
    m_tempAdcOnePV(nds::PVDelegateIn<int32_t>("TempADC-1", std::bind(&ADQDevice::getTempADCone, this,
                                                                     std::placeholders::_1, std::placeholders::_2))),
    m_tempAdcTwoPV(nds::PVDelegateIn<int32_t>("TempADC-2", std::bind(&ADQDevice::getTempADCtwo, this,
                                                                     std::placeholders::_1, std::placeholders::_2))),
    m_tempFpgaPV(nds::PVDelegateIn<int32_t>("TempFPGA", std::bind(&ADQDevice::getTempFPGA, this, std::placeholders::_1,
                                                                  std::placeholders::_2))),
    m_tempDCDC2APV(nds::PVDelegateIn<int32_t>("TempDCDC2A", std::bind(&ADQDevice::getTempDCDC2A, this, std::placeholders::_1,
                                                                      std::placeholders::_2))),
    m_tempDCDC2BPV(nds::PVDelegateIn<int32_t>("TempDCDC2B", std::bind(&ADQDevice::getTempDCDC2B, this, std::placeholders::_1,
                                                                      std::placeholders::_2))),
    m_tempDCDC1PV(nds::PVDelegateIn<int32_t>("TempDCDC1", std::bind(&ADQDevice::getTempDCDC1, this, std::placeholders::_1,
                                                                    std::placeholders::_2))),
    m_tempRSVDPV(nds::PVDelegateIn<int32_t>("TempRSVD", std::bind(&ADQDevice::getTempRSVD, this, std::placeholders::_1,
                                                                  std::placeholders::_2))),
    m_sampRatePV(nds::PVDelegateIn<double>("SampRate", std::bind(&ADQDevice::getSampRate, this, std::placeholders::_1,
                                                                 std::placeholders::_2))),
    m_bytesPerSampPV(nds::PVDelegateIn<int32_t>("BytesPerSample", std::bind(&ADQDevice::getBytesPerSample, this,
                                                                            std::placeholders::_1, std::placeholders::_2))),
    m_busTypePV(nds::PVDelegateIn<int32_t>("BusType", std::bind(&ADQDevice::getBusType, this, std::placeholders::_1,
                                                                std::placeholders::_2))),
    m_busAddrPV(nds::PVDelegateIn<int32_t>("BusAddr", std::bind(&ADQDevice::getBusAddr, this, std::placeholders::_1,
                                                                std::placeholders::_2))),
    m_pcieLinkRatePV(nds::PVDelegateIn<int32_t>("PCIeLinkRate", std::bind(&ADQDevice::getPCIeLinkRate, this,
                                                                          std::placeholders::_1, std::placeholders::_2))),
    m_pcieLinkWidPV(nds::PVDelegateIn<int32_t>("PCIeLinkWid", std::bind(&ADQDevice::getPCIeLinkWid, this,
                                                                        std::placeholders::_1, std::placeholders::_2))),
    m_node(nds::Port(name + INFO_DEVICE, nds::nodeType_t::generic)),
    m_adqInterface(adqInterface),
    m_sampRateDecPV(nds::PVDelegateIn<double>("SampRateDec", std::bind(&ADQDevice::getSampRateDec, this,
                                                                       std::placeholders::_1, std::placeholders::_2)))
{
    parentNode.addChild(m_node);
    m_doHardwareMonitor = true;
    m_tempLocal = m_tempAdcOne = m_tempAdcTwo = m_tempFpga = m_tempDCDC2A = m_tempDCDC2B = m_tempDCDC1 = m_tempRSVD = 0;

    // PVs for device info
    m_productNamePV.setMaxElements(STRING_ENUM);
    m_productNamePV.processAtInit(PINI);
    m_node.addChild(m_productNamePV);

    m_serialNumberPV.setMaxElements(STRING_ENUM);
    m_serialNumberPV.processAtInit(PINI);
    m_node.addChild(m_serialNumberPV);

    m_productIDPV.processAtInit(PINI);
    m_node.addChild(m_productIDPV);

    m_adqTypePV.processAtInit(PINI);
    m_node.addChild(m_adqTypePV);

    m_cardOptionPV.processAtInit(PINI);
    m_cardOptionPV.setMaxElements(STRING_ENUM);
    m_node.addChild(m_cardOptionPV);

    // PVs for temperatures
    nds::enumerationStrings_t isAliveList = { "Fail", "OK"};
    m_isAlivePV.setEnumeration(isAliveList);
    m_isAlivePV.setScanType(nds::scanType_t::periodic);
    m_isAlivePV.processAtInit(PINI);
    m_node.addChild(m_isAlivePV);
    m_isAlive = true;

    m_tempLocalPV.processAtInit(PINI);
    m_node.addChild(m_tempLocalPV);

    m_tempAdcOnePV.processAtInit(PINI);
    m_node.addChild(m_tempAdcOnePV);

    m_tempAdcTwoPV.processAtInit(PINI);
    m_node.addChild(m_tempAdcTwoPV);

    m_tempFpgaPV.processAtInit(PINI);
    m_node.addChild(m_tempFpgaPV);

    m_tempDCDC2APV.processAtInit(PINI);
    m_node.addChild(m_tempDCDC2APV);

    m_tempDCDC2BPV.processAtInit(PINI);
    m_node.addChild(m_tempDCDC2BPV);

    m_tempDCDC1PV.processAtInit(PINI);
    m_node.addChild(m_tempDCDC1PV);

    m_tempRSVDPV.processAtInit(PINI);
    m_node.addChild(m_tempRSVDPV);

    // PV for sample rate
    m_sampRatePV.processAtInit(PINI);
    m_node.addChild(m_sampRatePV);

    m_sampRatePV.processAtInit(PINI);
    m_node.addChild(m_sampRateDecPV);

    // PV for number of bytes per sample
    m_bytesPerSampPV.processAtInit(PINI);
    m_node.addChild(m_bytesPerSampPV);

    // PV for Bus connection
    m_busAddrPV.processAtInit(PINI);
    m_node.addChild(m_busAddrPV);

    m_busTypePV.processAtInit(PINI);
    m_node.addChild(m_busTypePV);

    m_pcieLinkRatePV.processAtInit(PINI);
    m_node.addChild(m_pcieLinkRatePV);

    m_pcieLinkWidPV.processAtInit(PINI);
    m_node.addChild(m_pcieLinkWidPV);
}

void ADQDevice::getProductName(timespec* pTimestamp, std::string* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    *pValue = m_adqInterface->GetBoardProductName();
    *pTimestamp = m_productNamePV.getTimestamp();
}

void ADQDevice::getSerialNumber(timespec* pTimestamp, std::string* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    *pValue = m_adqInterface->GetBoardSerialNumber();
    *pTimestamp = m_serialNumberPV.getTimestamp();
}

void ADQDevice::getProductID(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    *pValue = m_adqInterface->GetProductID();
    *pTimestamp = m_productIDPV.getTimestamp();
}

void ADQDevice::getADQType(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    *pValue = m_adqInterface->GetADQType();
    *pTimestamp = m_adqTypePV.getTimestamp();
}

void ADQDevice::getCardOption(timespec* pTimestamp, std::string* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    *pValue = m_adqInterface->GetCardOption();
    *pTimestamp = m_cardOptionPV.getTimestamp();
}

void ADQDevice::isAlive(timespec* pTimestamp, int32_t* pValue)
{
    if (!m_adqInterface)
        m_isAlive = false;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        m_isAlive = m_adqInterface->IsAlive();
    }
    *pValue = m_isAlive;
    *pTimestamp = m_isAlivePV.getTimestamp();
}

void ADQDevice::getTempLocal(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempLocal = m_adqInterface->GetTemperature(TEMP_LOCAL) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempLocal;
    *pTimestamp = m_tempLocalPV.getTimestamp();
}

void ADQDevice::getTempADCone(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempAdcOne = m_adqInterface->GetTemperature(TEMP_ADC_ONE) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempAdcOne;
    *pTimestamp = m_tempAdcOnePV.getTimestamp();
}

void ADQDevice::getTempADCtwo(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempAdcTwo = m_adqInterface->GetTemperature(TEMP_ADC_TWO) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempAdcTwo;
    *pTimestamp = m_tempAdcTwoPV.getTimestamp();
}

void ADQDevice::getTempFPGA(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempFpga = m_adqInterface->GetTemperature(TEMP_FPGA) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempFpga;
    *pTimestamp = m_tempFpgaPV.getTimestamp();
}

void ADQDevice::getTempDCDC2A(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempDCDC2A = m_adqInterface->GetTemperature(TEMP_DCDC2A) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempDCDC2A;
    *pTimestamp = m_tempDCDC2APV.getTimestamp();
}

void ADQDevice::getTempDCDC2B(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempDCDC2B = m_adqInterface->GetTemperature(TEMP_DCDC2B) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempDCDC2B;
    *pTimestamp = m_tempDCDC2BPV.getTimestamp();
}

void ADQDevice::getTempDCDC1(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempDCDC1 = m_adqInterface->GetTemperature(TEMP_DCDC1) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempDCDC1;
    *pTimestamp = m_tempDCDC1PV.getTimestamp();
}

void ADQDevice::getTempRSVD(timespec* pTimestamp, int32_t* pValue)
{
    if ((!m_adqInterface) || (!m_isAlive))
        *pValue = 0;
    else if (m_doHardwareMonitor)
    {
        std::lock_guard<std::mutex> lock(m_adqDevMutex);
        *pValue = m_tempRSVD = m_adqInterface->GetTemperature(TEMP_RSVD) * CELSIUS_CONVERT;
    }
    else
        *pValue = m_tempRSVD;
    *pTimestamp = m_tempRSVDPV.getTimestamp();
}

void ADQDevice::getSampRate(timespec* pTimestamp, double* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    double sampRate = 0;
    if (!m_adqInterface)
        return;
    m_adqInterface->GetSampleRate(0, &sampRate);
    *pValue = sampRate;
    *pTimestamp = m_sampRatePV.getTimestamp();
}

void ADQDevice::getSampRateDec(timespec* pTimestamp, double* pValue)
{
    if (!m_adqInterface)
        return;
    double sampRateDec = 0;
    m_adqInterface->GetSampleRate(1, &sampRateDec);
    *pValue = sampRateDec;
    *pTimestamp = m_sampRateDecPV.getTimestamp();
}

void ADQDevice::getBytesPerSample(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    unsigned int bytes = 0;
    if (!m_adqInterface)
        return;
    m_adqInterface->GetNofBytesPerSample(&bytes);
    *pValue = bytes;
    *pTimestamp = m_bytesPerSampPV.getTimestamp();
}

void ADQDevice::getBusAddr(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    if ((m_adqInterface->IsPCIeDevice()) || (m_adqInterface->IsPCIeLiteDevice()))
    {
        *pValue = m_adqInterface->GetPCIeAddress();
    }

    if ((m_adqInterface->IsUSBDevice()) || (m_adqInterface->IsUSB3Device()))
    {
        *pValue = m_adqInterface->GetUSBAddress();
    }

    *pTimestamp = m_busAddrPV.getTimestamp();
}

void ADQDevice::getBusType(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    if (m_adqInterface->IsPCIeDevice())
    {
        *pValue = 0;
    }

    if (m_adqInterface->IsPCIeLiteDevice())
    {
        *pValue = 1;
    }

    if (m_adqInterface->IsUSBDevice())
    {
        *pValue = 2;
    }

    if (m_adqInterface->IsUSB3Device())
    {
        *pValue = 3;
    }

    *pTimestamp = m_busTypePV.getTimestamp();
}

void ADQDevice::getPCIeLinkRate(timespec* pTimestamp, int32_t* pValue)
{
    std::lock_guard<std::mutex> lock(m_adqDevMutex);
    if (!m_adqInterface)
        return;
    if ((m_adqInterface->IsPCIeDevice()) || (m_adqInterface->IsPCIeLiteDevice()))
    {
        *pValue = m_adqInterface->GetPCIeLinkRate();
    }

    *pTimestamp = m_pcieLinkRatePV.getTimestamp();
}

void ADQDevice::getPCIeLinkWid(timespec* pTimestamp, int32_t* pValue)
{
    if (!m_adqInterface)
        return;
    if ((m_adqInterface->IsPCIeDevice()) || (m_adqInterface->IsPCIeLiteDevice()))
    {
        *pValue = m_adqInterface->GetPCIeLinkWidth();
    }

    *pTimestamp = m_pcieLinkWidPV.getTimestamp();
}

ADQDevice::~ADQDevice()
{
    m_adqInterface = NULL;
}
